--
-- 2023.06.13: add svec3.sub

local tpftransf = require "transf"
local tpfvec3 = require "vec3"


local vec3 = table.copy(tpfvec3)
local transf = table.copy(tpftransf)

local svec3arr = {}
local svec3 = { arr = svec3arr }


function vec3.toArray(v)
	return { v.x, v.y, v.z }
end

function vec3.addArray(v, b)
	local result = {
		x = v.x + b[1],
		y = v.y + b[2],
		z = v.z + b[3]
	}
	return result
end

function vec3.fromArray(a)
	return { x = a[1], y = a[2], z = a[3] }
end

function vec3.cross(a, b)
	local result = {
		x = a.y * b.z - a.z * b.y,
		y = a.z * b.x - a.x * b.z,
		z = a.x * b.y - a.y * b.x
	}
	return result
end

function vec3.transformByMatrix(v, trans)
	local result = vec3.new( 
		(v.x * trans[1]) + (v.y * trans[5]) + (v.z * trans[9]) + trans[13],
		(v.x * trans[2]) + (v.y * trans[6]) + (v.z * trans[10]) + trans[14],
		(v.x * trans[3]) + (v.y * trans[7]) + (v.z * trans[11]) + trans[15]
	)
	return result
end

function svec3.transformByMatrix(v, trans)
	local result = {
		(v[1] * trans[1]) + (v[2] * trans[5]) + (v[3] * trans[9]) + trans[13],
		(v[1] * trans[2]) + (v[2] * trans[6]) + (v[3] * trans[10]) + trans[14],
		(v[1] * trans[3]) + (v[2] * trans[7]) + (v[3] * trans[11]) + trans[15]
	}
	return result
end


function vec3.rotateXRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = v.x,
		y = (v.y * c - v.z * s),
		z = (v.y * s + v.z * c)
	}
	return result
end

function vec3.rotateX(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = v.x,
		y = (v.y * c - v.z * s),
		z = (v.y * s + v.z * c)
	}
	return result
end


function svec3.rotateXRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		v[1], 
		(v[2] * c - v[3] * s),
		(v[2] * s + v[3] * c)
	}
	return result
end

function svec3.rotateX(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		v[1], 
		(v[2] * c - v[3] * s),
		(v[2] * s + v[3] * c)
	}
	return result
end


function vec3.rotateYRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = (v.x * c + v.z * s),
		y = y,
		z = (v.x * -s + v.z * c)
	}
	return result
end

function vec3.rotateY(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = (v.x * c + v.z * s),
		y = y,
		z = (v.x * -s + v.z * c)
	}
	return result
end


function svec3.rotateYRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		(v[1] * c + v[3] * s),
		v[2],
		(v[1] * -s + v[3] * c)
	}
	return result
end

function svec3.rotateY(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		(v[1] * c + v[3] * s),
		v[2],
		(v[1] * -s + v[3] * c)
	}
	return result
end


function vec3.rotateZRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = (v.x * c - v.y * s),
		y = (v.x * s + v.y * c),
		z = v.z
	}
	return result
end

function vec3.rotateZ(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		x = (v.x * c - v.y * s),
		y = (v.x * s + v.y * c),
		z = v.z
	}
	return result
end

function svec3.rotateZRad(v, rot)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		(v[1] * c - v[2] * s),
		(v[1] * s + v[2] * c),
		v[3]
	}
	return result
end

function svec3.rotateZ(v, d)
	local rot = math.rad(d)
	local s = math.sin(rot)
	local c = math.cos(rot)

	local result = {
		(v[1] * c - v[2] * s),
		(v[1] * s + v[2] * c),
		v[3]
	}
	return result
end


function svec3.add(a, b)
	local result = {
		a[1] + b[1],
		a[2] + b[2],
		a[3] + b[3]
	}
	return result
end

function svec3.sub(a, b)
	local result = {
		a[1] - b[1],
		a[2] - b[2],
		a[3] - b[3]
	}
	return result
end

function svec3.mul(f, v)
	local result = {
		v[1]*f,
		v[2]*f,
		v[3]*f
	}
	return result
end


function svec3.dot(a, b)
	return a[1] * b[1] + a[2] * b[2] + a[3] * b[3];
end

function svec3.cross(a, b)
	local result = {
		a[2] * b[3] - a[3] * b[2],
		a[3] * b[1] - a[1] * b[3],
		a[1] * b[2] - a[2] * b[1]
	}
	return result
end

function svec3.length(v)
	return math.sqrt(v[1]^2 + v[2]^2 + v[3]^2)
end

function svec3.lengthBetween(a, b)
	return math.sqrt( (a[1]-b[1])^2 + (a[2]-b[2])^2 + (a[2]-b[2])^2)
end

function svec3.normalize(v)
	return svec3.mul(1.0 / svec3.length(v), v)
end

function svec3.angleUnit(a, b)
	arg = svec3.dot(a, b)
	
	if (arg < -1.0) then 
		arg = -1.0
	elseif (arg > 1.0) then 
		arg = 1.0
	end
	
	return math.acos(arg)
end

function svec3.xyAngle(a)
	return math.atan2(a[2], a[1])
end

function svec3arr.add(arr, b)
	local result = {}
	for i = 1, #arr do
		result[i] = svec3.add(arr[i], b)
	end
	return result
end

function svec3arr.addVec3(arr, v)
	local result = { }
	local b = { v.x, v.y, v.z }
	for i = 1, #arr do
		result[i] = svec3.add(arr[i], b)
	end
	return result
end

function svec3arr.rotateX(arr, d)
	local result = {}
	for i = 1, #arr do
		result[i] = svec3.rotateX(arr[i], d)
	end
	return result
end

function svec3arr.rotateY(arr, d)
	local result = {}
	for i = 1, #arr do
		result[i] = svec3.rotateY(arr[i], d)
	end
	return result
end

function svec3arr.rotateZ(arr, d)
	local result = {}
	for i = 1, #arr do
		result[i] = svec3.rotateZ(arr[i], d)
	end
	return result
end

function svec3arr.transformByMatrix(arr, trans)
	local result = {}
	for i = 1, #arr do
		result[i] = svec3.transformByMatrix(arr[i], trans)
	end
	return result
end



function transf.decompositeParts(m)
	local scaleing = {}
	local transl = { m[13], m[14], m[15] }
	local rotatingMatrix = {}
	print( math.sqrt(m[1]^2 + m[2]^2 + m[3]^2) )
	scaleing[1] = m[16] * math.sqrt(m[1]^2 + m[2]^2 + m[3]^2)
	scaleing[2] = m[16] * math.sqrt(m[5]^2 + m[6]^2 + m[7]^2)
	scaleing[3] = m[16] * math.sqrt(m[9]^2 + m[10]^2 + m[11]^2)
	

	rotatingMatrix = {
		m[1] / scaleing[1], m[2]/scaleing[2], m[3] / scaleing[3], 0,
		m[5] / scaleing[1], m[6]/scaleing[2], m[7] / scaleing[3], 0,
		m[9] / scaleing[1], m[10]/scaleing[2], m[11] / scaleing[3], 0,
		0, 0, 0, 1
	}
	
	local rotParts =  {
		math.atan2(rotatingMatrix[10],rotatingMatrix[11]);
		math.atan2(-rotatingMatrix[9], rotatingMatrix[11]);
		math.atan2(-rotatingMatrix[5], rotatingMatrix[6]);
	}


	local rotParts =  {
		math.atan2( rotatingMatrix[7], rotatingMatrix[11]);
		math.atan2(-rotatingMatrix[9], rotatingMatrix[11]);
		math.atan2(-rotatingMatrix[5], rotatingMatrix[6]);
	}


	local rotParts =  {
		math.atan2( rotatingMatrix[7], rotatingMatrix[11]);
		math.atan2(-rotatingMatrix[9], rotatingMatrix[11]);
		math.atan2(rotatingMatrix[5], rotatingMatrix[1]);
	}




	local theta1 = math.atan2(m[7], m[11])
	local c2 = math.sqrt(m[1]^2 + m[2]^2)
	local theta2 = math.atan2(-m[3], c2)
	local s1 = math.sin(theta1)
	local c1 = math.cos(theta1)
	local theta3 = math.atan2(s1*m[9] - c1*m[5], c1*m[6] - s1*m[10])
	local rotParts = { -theta1, -theta2, -theta3 }

--   m[1]  m[2]  m[3]  m[4]
--   m[5]  m[6]  m[7]  m[8]
--   m[9] m[10] m[11] m[12]
--  m[13] m[14] m[15] m[16]

--[[


	if (m[5] > 0.998) then
		rotParts[1]  = math.atan2(m[3],m[11])
		rotParts[2] = math.pi/2
		rotParts[3] = 0
	elseif (m[5] < -0.998) then
		rotParts[1] = math.atan2 (m[3],m[11])
		rotParts[2] = -math.pi/2;
		rotParts[3] = 0
	else
		rotParts[1]  = math.atan2(-m[9],m[1])
		rotParts[3] = math.atan2(-m[7],m[6])
		rotParts[2] = math.asin(m[5])
	end
--]]
--[[
	local rotParts = {}

	local clamped = math.max(-1, math.min(1, m[11]))
	rotParts[2] = math.asin(clamped);

	if (math.abs( m[11]) < 0.99999) then
		rotParts[1] = math.atan2(-m[13], m[14]);
		rotParts[3] = math.atan2(-m[6], m[1]);
	else
		rotParts[1] = math.atan2(m[9], m[7]);
		rotParts[3] = 0
	end
--]]
	local rotPartsDeg = {
		math.deg(rotParts[1]),
		math.deg(rotParts[2]),
		math.deg(rotParts[3]),
	}

	local data = {
		transl = transl,
		scaleing = scaleing,
		rotatingMatrix = rotatingMatrix,
		rotationAngles = rotParts,
		rotPartsDeg = rotPartsDeg,
	}
	return data
end

local mathmisc = {
	vec3 = vec3,
	svec3 = svec3,
	transf = transf,
}
return mathmisc
