local checkModels = {}

local _utils = commonapi.utils
local _log = _utils.logPrep("commonapi2.checkModels")


local _syncdata = commonapi._syncdata


local loaderrors_sync =
	_syncdata.SyncObject:new(
	{
		key = "loaderrors",
		data = {}
	}
)
loaderrors_sync:useFailbackIfRequired()
loaderrors_sync:setResetOnGameLoad()

local datastorage = loaderrors_sync:getData({})

function checkModels.getErrors(failure)
	datastorage = loaderrors_sync:getData()
	return datastorage.models or {}
end

function checkModels.resetErrors(failure)
	_log("checkModels.resetErrors")
	datastorage = {}
	loaderrors_sync:syncData(datastorage)
end

function checkModels.addErrorData(fileName, data)
	if (data.text ~= nil) then
		print("CommonAPI2 CheckModels ERROR: "..fileName..": "..data.text)
	end
	datastorage = loaderrors_sync:getData()
	if (type(datastorage.models) ~= "table") then
		datastorage.models = {}
	end

	local dmods = datastorage.models
	if (type(dmods[fileName]) ~= "table") then
		dmods[fileName] = {}
	end

	local entry = dmods[fileName]
	if (type(entry.lines) ~= "table") then
		entry.lines = {}
		entry.maxseverity = 0
	end
	local lines = entry.lines
	lines[#lines+1] = data

	local dataSeverity = data.severity or 0
	if (dataSeverity > entry.maxseverity) then
		entry.maxseverity = dataSeverity
	end
	loaderrors_sync:syncData()
end


local function calculateMaxChildren(lodData)
	if (lodData == nil) then
		return 0
	end
	local c = 1
	if (lodData.children ~= nil) then
		for i = 1, #lodData.children do
			c = c + calculateMaxChildren(lodData.children[i])
		end
	end
	return c
end


local lodChecker = {}

function lodChecker:getTableByPath(data, parts, basepath)
	local ptr = data
	for idx = 1, #parts do
		local key = parts[idx]
		local newptr = ptr[key]
		if (newptr == nil) then
			return nil
		end
		if (type(newptr) ~= "table") then
			local errpath = basepath or ""
			for i = 1, idx do
				errpath = errpath .. "."..parts[i]
			end
			self:addErrorAt(errpath, "should be a table?!")
			return nil
		end
		ptr = newptr
	end
	return ptr
end


function lodChecker:insertError(data)
	checkModels.addErrorData(self.fileName, data)
end

function lodChecker:addError(failure)
	self:insertError({
		text = failure
	})
end

function lodChecker:addErrorAt(metadatapath, failure, severity)
	local lodaccess = ""
	local lodId = self.lodId or nil
	if (lodId == nil) then
		lodaccess = " lodid = ".. lodId .." (LOD"..(lodId-1)..")"
	end
	local d = {
		lodId = lodId,
		metadatapath = metadatapath,
		failure = failure,
		text = metadatapath.." "..failure..lodaccess
	}
	if (severity ~= nil) then
		d.severity = severity
	end
	self:insertError(d)
end

function lodChecker:setCurrentLod(lodId)
	if (lodId < 0) then
		self.lodId = nil
		self.lodNode = nil
		self.meshIdMax = -1
	end
	self.lodId = lodId
	local lod = self.data.lods[self.lodId]
	if (type(lod) ~= "table" or type(lod.node) ~= "table") then
		self.lodNode = nil
		self.meshIdMax = -1
	else
		self.lodNode = lod.node
		self.meshIdMax = calculateMaxChildren(self.lodNode) - 1
	end
end

function lodChecker:checkMeshIdInBounds(keypath, meshId)
	if (self.meshIdMax < 0) then
		self:addErrorAt(keypath, " access non existent lod!")
		return false
	end
	if (meshId > self.meshIdMax) then
		self:addErrorAt(keypath, " meshId = "..meshId .." bigger then last mesh id in lod = "..self.meshIdMax, 100)
		return false
	end
	return true
end


function lodChecker:addConfigsError(configpath, failure, severity)
	local lodId = self.lodId
	local lodName = "(LOD"..(lodId-1)..")"
	local metadatapath = self.metadataFeature..".configs["..self.configIdx.."]."..configpath
	local lodaccess = " lodid = ".. lodId .." "..lodName.."."

	local d = {
		lodId = lodId,
		metadatapath = metadatapath,
		failure = failure,
		text = metadatapath.." "..failure..lodaccess
	}
	if (severity ~= nil) then
		d.severity = severity
	end
	self:insertError(d)
end


function lodChecker:checkConfigsIn(feature, fn)
	if (type(self.data.metadata) ~= "table") then
		self:addError("metadata should be a table")
		return
	end

	local entry = self.data.metadata[feature]
	if (type(entry) ~= "table") then
		self:addErrorAt(feature, "should be a table")
		return
	end

	local configs = entry.configs
	if (type(configs) ~= "table") then
		self:addErrorAt(feature..".configs", "should be a table")
		return
	end

	if (type(self.data.lods) ~= "table") then
		self:addError("access non existent lods table!")
		return false
	end

	for configIdx = 1, #configs do
		self.metadataFeature = feature
		self.configIdx = configIdx
		self.config = configs[configIdx]

		local lodId = configIdx
		self:setCurrentLod(lodId)

		self.meshIdsUsed = {}
		fn(self)
	end

	self.metadataFeature = nil
	self.configIdx = nil
	self.config = nil
	self:setCurrentLod(-1)
end

function lodChecker:checkConfig(configname, alternative, disallowZero)
	if (self.config[configname] == nil) then
		return
	end
	local c = self.config[configname]
	if (type(c) ~= "table") then
		self:addConfigsError(configname, " should be table?!")
		return
	end

	for i = 1, #c do
		local meshId = c[i]
		local configpath = configname.. "["..i.."]";

		if (type(meshId) ~= "number") then
			self:addConfigsError(configpath, " not a number?!")
		else
			--print(self.fileName .. " meshId = "..meshId.. " last mesh id = "..self.meshIdMax)
			if (disallowZero == true and meshId == 0) then
				self:addConfigsError(configpath, " meshId = "..meshId .." looks like an error", 5)
			end
			if (self:checkConfigMeshIdInBounds(configpath, meshId)) then
				if (self.meshIdsUsed[meshId] ~= nil) then
					local lastUsage = self.meshIdsUsed[meshId]
					if ( (lastUsage == configname) or (lastUsage == alternative) ) then
						-- duplicate usage
					else
						self:addConfigsError(configpath, " meshId = "..meshId .." already used by "..self.meshIdsUsed[meshId], 10)
					end
				end
			end
			self.meshIdsUsed[meshId] = configname
		end
	end
end

function lodChecker:checkConfigMeshIds(configname)
	if (self.config[configname] == nil) then
		return
	end
	local c = self.config[configname]
	if (type(c) ~= "table") then
		self:addConfigsError(configname, " should be table?!")
		return
	end

	for i = 1, #c do
		local meshId = c[i]
		local configpath = configname.. "["..i.."]";

		if (type(meshId) ~= "number") then
			self:addConfigsError(configpath, " not a number?!")
		else
			self:checkConfigMeshIdInBounds(configpath, meshId)
		end
	end
end


function lodChecker:checkConfigMeshIdInBounds(configpath, meshId)
	if (self.meshIdMax < 0) then
		self:addConfigsError(configpath, " access non existent lod!", 100)
		return false
	end
	if (meshId > self.meshIdMax) then
		self:addConfigsError(configpath, " meshId = "..meshId .." bigger then last mesh id in lod = "..self.meshIdMax, 100)
		return false
	end
	return true
end



function lodChecker:checkConfigFakeBogies()
	local configname = "fakeBogies"
	if (self.config[configname] == nil) then
		return
	end
	local c = self.config[configname]
	if (type(c) ~= "table") then
		self:addConfigsError(configname," should be table?!")
		return
	end
	for i = 1, #c do
		local configpath = configname.. "["..i.."]";
		local fakeBogie = c[i]

		if (type(fakeBogie) ~= "table") then
			self:addConfigsError(configpath, " should be a table?!")
		else
			configpath = configpath..".group";
			if (type(fakeBogie.group) ~= "number") then
				self:addConfigsError(configpath, "should be a number?!, currently is "..type(fakeBogie.group) )
			else
				local meshId = fakeBogie.group
			--[[ Extra check as mentioned by VacuumTube
			if (meshId == 0) then
				local offsetSet = false
				local positionSet = false
				if (fakeBogie.offset ~= nil and (type(fakeBogie.offset) == "number") and fakeBogie.offset ~= 0) then
					offsetSet = true
				end
				if (fakeBogie.position ~= nil and (type(fakeBogie.position) == "number") and fakeBogie.position ~= 0) then
					positionSet = true
				end

				if (positionSet == false and offsetSet == false) then
					if ((type(self.config.axles) == "table") and #self.config.axles == 0) then
						self:addConfigsError(configpath, "meshId = 0 and #config[].axles == 0 seems to be incorrect")
					end
				end
			end
			--]]
				self:checkConfigMeshIdInBounds(configpath, meshId)
			end  -- check type(fakeBogie.group)
		end
	end
end



function lodChecker:checkConfigWithIds(configname)
	if (self.config[configname] == nil) then
		return
	end
	local data = self.config[configname]
	if (type(data) ~= "table") then
		self:addConfigsError(configname," should be table?!")
		return
	end
	local configpath = configname..".ids";
	if (type(data.ids) ~= "table") then
		self:addConfigsError(configpath, "should be a table?!")
	end
	for i, meshId in pairs(data.ids) do
		self:checkConfigMeshIdInBounds(configpath.."["..i.."]", meshId)
	end
end


function lodChecker:checkSeatProvider()
	if (self.data.metadata == nil) then
		return
	end
	if (self.data.metadata.seatProvider == nil) then
		return
	end
	local seatProvider = self.data.metadata.seatProvider
	if (seatProvider.seats == nil) then
		return
	end

	self:setCurrentLod(1)
	if (type(seatProvider.seats) ~= "table") then
		self:addErrorAt("seatProvider.seats", "should be a table?!")
	end
	for i, seat in ipairs(seatProvider.seats) do
		local keypath = "seatProvider.seats["..i.."]"

		if (type(seat) ~= "table") then
			self:addErrorAt(keypath, "should be a table?!")
		else
			keypath = keypath..".group";
			if (type(seat.group) ~= "number") then
				self:addErrorAt(keypath, "should be a number?!")
			end
			self:checkMeshIdInBounds(keypath, seat.group)
		end
	end
	self:setCurrentLod(-1)
end



function lodChecker:checkCargoSlotProvider()
	if (self.data.metadata == nil) then
		return
	end
	if (self.data.metadata.cargoSlotProvider == nil) then
		return
	end
	local cargoSlotProvider = self.data.metadata.cargoSlotProvider
	if (cargoSlotProvider.slots == nil) then
		return
	end

	self:setCurrentLod(1)
	if (type(cargoSlotProvider.slots) ~= "table") then
		self:addErrorAt("cargoSlotProvider.slots", "should be a table?!")
	end
	for i, seat in ipairs(cargoSlotProvider.slots) do
		local keypath = "cargoSlotProvider.slots["..i.."]"

		if (type(seat) ~= "table") then
			self:addErrorAt(keypath, " should be a table?!")
		else
			keypath = keypath..".group";
			if (type(seat.group) ~= "number") then
				self:addErrorAt(keypath, "should be a number?!")
			end
			self:checkMeshIdInBounds(keypath, seat.group)
		end
	end
	self:setCurrentLod(-1)
end

function lodChecker:checkSeats(cargoEntry, cargoEntryPath, maxSeatsTPF)
	if (cargoEntry.seats == nil) then
		return
	end
	if (type(cargoEntry.seats) ~= "table") then
		self:addErrorAt(cargoEntryPath..".seats", "should be a table?!")
		return
	end
	if (#cargoEntry.seats == 0) then
		-- having no seats defined seems to be ok
		return
	end

	if (maxSeatsTPF > -1) then
		for seatIdx, seatId in ipairs(cargoEntry.seats) do
			if (type(seatId) == "number") then
				if (seatId < 0) then
					self:addErrorAt(cargoEntryPath..".seats["..seatIdx.."]", " = "..seatId..", but needs to be 0 or greater,  valid range: 0 - "..maxSeatsTPF)
				elseif (seatId > maxSeatsTPF) then
					self:addErrorAt(cargoEntryPath..".seats["..seatIdx.."]", " = "..seatId..", is greater then seatProviders, valid range: 0 - "..maxSeatsTPF, 100)
				end
			end
		end
	else
		self:addErrorAt(cargoEntryPath..".seats", "seatProviders is missing, can't use seats!")
	end
end

function lodChecker:checkLoadConfig(loadConfig, loadConfigPath, maxSeatsTPF)
	if (type(loadConfig) == "table") then
		if (type(loadConfig.cargoEntries) == "table") then
			local cargoEntries = loadConfig.cargoEntries
			if (type(cargoEntries) ~= "table") then
				self:addErrorAt(loadConfigPath..".cargoEntries", "should be a table?!")
				return
			end
			for cargoEntryIdx, cargoEntry in ipairs(cargoEntries) do
				local cargoEntryPath = loadConfigPath..".cargoEntries["..cargoEntryIdx.."]"
				if (type(cargoEntry) == "table") then
					self:checkSeats(cargoEntry, cargoEntryPath, maxSeatsTPF);
				else
					self:addErrorAt(cargoEntryPath, "should be a table?!")
				end
			end
		else
			if (loadConfig.cargoEntries ~= nil) then
				self:addErrorAt(loadConfigPath..".cargoEntries", "should be a table?!")
			end
		end
	else
		self:addErrorAt(loadConfigPath, "should be a table?!")
	end
end

function lodChecker:checkCompartmentsList()
	if (self.data.metadata == nil) then
		return
	end
	local compartmentsList = self:getTableByPath(self.data.metadata, { "transportVehicle", "compartmentsList" })
	if (compartmentsList == nil) then
		return
	end

	local maxSeats = -1

	local seats = self:getTableByPath(self.data.metadata, { "seatProvider", "seats" })
	if (seats ~= nil) then
		maxSeats = #seats
	end
	local maxSeatsTPF = maxSeats - 1

	for compartmentIdx, compartment in ipairs(compartmentsList) do
		if (compartment.loadConfigs ~= nil) then
			local compartmentPathLoadConfigs = "transportVehicle.compartmentsList["..compartmentIdx.."].loadConfigs"
			if (type(compartment.loadConfigs) == "table") then
				local loadConfigs = compartment.loadConfigs
				for loadConfigIdx, loadConfig in ipairs(loadConfigs) do
					local loadConfigPath = compartmentPathLoadConfigs.."["..loadConfigIdx.."]"
					self:checkLoadConfig(loadConfig, loadConfigPath, maxSeatsTPF)
				end
			else
				self:addErrorAt(compartmentPathLoadConfigs, "should be a table?!")
			end
		end
	end
end


function lodChecker.create(fileName, data)
	local modidRegex = "([^%/]+)/([^%/]+)(/res/models/model/.*%.mdl)$"
	local folder, modid, partfile = string.match(fileName, modidRegex)
	if (folder == "mods" or folder == "1066780" or folder == "staging_area") then
		fileName = modid..partfile
	end

	local o = {
		modid = modid,
		fileName = fileName,
		data = data,
		lodId = -1,
	}
	setmetatable(o, {
		__index = lodChecker
	})
	--print("fileName: ", fileName.."  "..lodId.." meshIdMax= "..o.meshIdMax)
	return o;
end

function checkModels.checkModelFile(fileName, model)
	if (model.metadata == nil) then
		return
	end
	if (model.lods == nil) then
		return
	end

	local metadata = model.metadata
	if (type(metadata) ~= "table") then
		--checkModels.addError(fileName, "metadata should be a table")
		return
	end
	local checker
	if (metadata.roadVehicle ~= nil) then
		checker = lodChecker.create(fileName, model)
		checker:checkConfigsIn("roadVehicle", function(o)
				o:checkConfig("blinkLightsLeft", nil, true)
				o:checkConfig("blinkLightsRight", nil, true)
				o:checkConfig("brakeLights", nil, true)
				o:checkConfig("headLights", nil, true)
				o:checkConfigFakeBogies()
				o:checkConfigMeshIds("steeringParts")
			end
		)
	end

	if (metadata.railVehicle ~= nil) then
		checker = lodChecker.create(fileName, model)
		checker:checkConfigsIn("railVehicle", function(o)
				o:checkConfig("frontForwardParts")
				o:checkConfig("innerForwardParts")
				o:checkConfig("backForwardParts")
				o:checkConfig("frontBackwardParts")
				o:checkConfig("innerBackwardParts")
				o:checkConfig("backBackwardParts")

				o:checkConfig("blinkingLights0", nil, true)
				o:checkConfig("blinkingLights1", nil, true)

				o:checkConfig("blinkLightsLeft0", "blinkingLights0", true)
				o:checkConfig("blinkLightsLeft1", "blinkingLights1", true)
				o:checkConfig("blinkLightsRight0", "blinkingLights0", true)
				o:checkConfig("blinkLightsRight1", "blinkingLights1", true)
				o:checkConfigFakeBogies()
			end
		)
	end


	if (metadata.airVehicle ~= nil) then
		checker = lodChecker.create(fileName, model)
		checker:checkConfigsIn("airVehicle", function(o)
				o:checkConfig("beaconLights", nil, true)
				o:checkConfig("landingLight", nil, true)
				o:checkConfig("strobeLights", nil, true)
				o:checkConfig("props", "propsB")

				o:checkConfigMeshIds("steeringParts")
				o:checkConfigFakeBogies()

				o:checkConfigWithIds("aileronLeft")
				o:checkConfigWithIds("aileronRight")
				o:checkConfigWithIds("elevator")
				o:checkConfigWithIds("flaps")
				o:checkConfigWithIds("rudder")
			end
		)
	end

	if (metadata.waterVehicle ~= nil) then
		checker = lodChecker.create(fileName, model)
		checker:checkConfigsIn("waterVehicle", function(o)
				o:checkConfigWithIds("paddles")
				o:checkConfigWithIds("rudder")
			end
		)
	end

	if (metadata.seatProvider ~= nil) then
		if (checker == nil) then
			checker = lodChecker.create(fileName, model)
		end
		checker:checkSeatProvider()
	end

	if (metadata.transportVehicle ~= nil) then
		if (checker == nil) then
			checker = lodChecker.create(fileName, model)
		end
		checker:checkCompartmentsList()
	end

	if (metadata.cargoSlotProvider ~= nil) then
		if (checker == nil) then
			checker = lodChecker.create(fileName, model)
		end
		checker:checkCargoSlotProvider()
	end
end

function checkModels.modifier(fileName, data)
	local ok, err = pcall(checkModels.checkModelFile, fileName, data)
	if (ok == false) then
		_log("checkModels failed for", fileName);
		local errcallok, errcallerr = pcall(checkModels.addErrorData, fileName, { text = err } )
		if (errcallok == false) then
			print(err)
			print(errcallerr)
		end
	end
	return data
end

checkModels._addModifer = function()
	addModifier("loadModel", checkModels.modifier)
end


return checkModels