--- Main common api and loader
-- @author Oskar Eisemuth
-- @copyright 2017-2024
-- @module commonapi

--assert(commonapi ~= nil, "Common API 2 init should be loaded once into lua")


local commonapi = {
--- Version of Common API as integer: YYYYMMDD
	version = 20240303,
	dmp = function(...)
		local ok, inspect = pcall(require, "commonapi2.extlib.inspect")
		if (ok) then
			print(inspect(...))
		end
	end,

	rerequire = function(name)
		package.loaded[name] = nil
		return require(name)
	end,
	__doc__ = "",
	__members = {},
}



local commonapi_internal = {
	_allowed = {
		["__commonapienv"] = true,
		["_events"] = true,
		["_modmm"] = true,
		["_nativeui"] = true,
		["_syncdata"] = true,
		["_shareddatafile"] = true,
		["_vehanimations"] = true,
		["filter"] = true,
		["mods"] = true,
		["repos"] = true,
		["settings"] = true,
		["stationhelper"] = true,
		["storage"] = true,
		["supports"] = true,
		["ui"] = true,
		["uiparameter"] = true,
		["utils"] = true,
		["nodedata"] = true,
	},
	_missing = {
		_native = true,
		-- ui = true,
	},
	_unique_threadid = tostring( {} ):sub(8),
	_unique_threadid_check = function()
		local ourthreadid = tostring( {} ):sub(8)
		if (commonapi_internal._unique_threadid == ourthreadid)  then
			print("commonapi2._unique_threadid_check() threadid match = "  ..ourthreadid)
		else
			print("commonapi2._unique_threadid_check() MISMATCH ".. commonapi_internal._unique_threadid .. " !="..ourthreadid)
		end
	end,
}


local commonapi_versionsdetect = {
	steam_35720_1 = true,
	steam_35720_2 = true,
	gog_35720_1 = true,

	steam_35731_1 = true,
	steam_35731_2 = true,
	gog_35731_1 = true,

	steam_35732_1 = true,
	steam_35732_2 = true,
	gog_35732_1 = true,
	gog_35732_2 = true,
}


commonapi._reloadModule = function(key)
	print("commonapi2._reloadModule "..key)
	rawset(commonapi, key, nil)
	rawset(commonapi_internal._missing, key, nil)
	package.loaded["commonapi2."..key] = nil
end

local commonapi_metatable = {
	__index = function(t, key)
		if (commonapi_internal._allowed[key] ~= true) then
			return nil
		end
		if (commonapi_internal._missing[key] ~= nil) then
			return nil
		end
		local ok, obj = pcall(require, "commonapi2."..key)
		if (ok) then
			if (type(obj) == "table") then
				t[key] = obj
				return t[key]
			else
				print("commonapi2 can't load "..key.." currently, native code not loaded?")
			end
		else
			print("commonapi2."..key.." not found")
		end
		commonapi_internal._missing[key] = true
		--print(debug.traceback())
		return nil
	end
}

setmetatable(commonapi, commonapi_metatable)

local function detectOS()
	local libext = package.cpath:match("%.[\\|/]?%?%.(%a+)")
	local mypath = debug.getinfo(2, "S").source:sub(2)

	local libexttosys = {
		dll = 1,
		so = 2,
		dylib = 3
	}

	mypath = string.gsub(mypath, '\\', '/')
	mypath = string.gsub(mypath, "/res/scripts/commonapi2/init%.lua$", "/")

	local info = {
		sys = libexttosys[libext] or 0,
		libext = libext,
		path = mypath,
		settings = {}
	}

	commonapi._info = info
end


local function loadSettingsDefaults()
	local settings = commonapi._info.settings
	if (settings.linedestination == nil) then
		settings.linedestination = true
	end
	if (settings.improveui == nil) then
		settings.improveui = 1
	end
end

local function loadSettings()
	local result, luafnSettings, err
	local filename = commonapi._info.path.."settings.lua"
	local env = {}
	setmetatable(env, {__index=_G})
	luafnSettings, err = loadfile(filename, "bt", env)

	if (luafnSettings == nil) then
		return
	end

	result = luafnSettings()
	if (type(result) ~= "table") then
		print("commonapi2.init: Result of "..filename.. " not a table?!")
		result = {}
	end

	if (result.linedestination == nil) then
		result.linedestination = true
	end
	commonapi._info.settings = result
end

local function loadCommonAPILuaExtension(mystrings)
	local sys = commonapi._info.sys

	if not ( sys == 1 or sys == 2 ) then
	--if not ( sys == 1 ) then
		return
	end

	if (CommonAPILuaStateID ~= nil) then
		-- CommonAPI own lua state, no need to check
	else
		if (getBuildVersion == nil or getBuildPrefix == nil) then
			print("commonapi2.init: getBuildVersion missing, please upgrade TPF2")
			return
		end
		local ugbuild = getBuildPrefix() .. '_'..getBuildVersion()..'_'..commonapi._info.sys


		if (type(commonapi._info.settings.buildoverwrite) == "string") then
			if (commonapi._info.settings.buildoverwrite ~= "") then
				local buildoverwrite = commonapi._info.settings.buildoverwrite:match('^%s*([a-z]+_%d+_%d)%s*$') or ''
				if (buildoverwrite ~= "") then
					commonapi_versionsdetect[buildoverwrite] = true
				else
					local errormsg = "ERROR: buildoverwrite format wrong, <provider>_<buildnr>_<osnum> please check your input: \"" .. commonapi._info.settings.buildoverwrite .. "\""
					print("commonapi2.init: "..errormsg)
					commonapi._info.nativeerrors = errormsg
				end
			end
		end

		if (commonapi_versionsdetect[ugbuild] ~= true) then
			local errormsg = "Your TPF2 version '" ..ugbuild.."' isn't listed as known good version, please update CommonAPI2"
			print("commonapi2.init: "..errormsg)
			commonapi._info.versionunknown = true
			return
		end
	end
	local mypath = commonapi._info.path.."bin/?."..commonapi._info.libext

	if ( sys == 1 ) then
		if (mypath:match("^mods/") ~= nil) then
			--print("Trying to change relative commonapi2 path to absolute")
			local fullgamepath = package.cpath:match("([^;]+\\)loadall%.dll;")
			if (fullgamepath ~= nil) then
			--	commonapi._info.path.."bin/?."..commonapi._info.libext
				local mynewpath = fullgamepath..string.gsub(mypath, '/', '\\')
				mypath = package.cpath..";"..mynewpath
			--	print("New cpath: ", mypath)
			end
		end
	end

	local oldcpath = package.cpath
	local oldpath = package.path
	package.cpath = mypath
--	package.path = ""	-- has to be a valid path, otherwise new game error at animals

	commonapi_internal._missing['_native'] = true
	--commonapi_internal._missing['ui'] = true
	local ok, dll = pcall(require, "CommonAPI2Native")
	package.cpath = oldcpath
	package.path = oldpath

	if not ok then
		local errormsg = "Can't load native code CommonAPI2Native: "..dll
		commonapi._info.nativeerrors = errormsg
		print("commonapi2.init: "..errormsg)
		return
	elseif (type(dll) == 'table') then
		commonapi._native = dll

		--print("commonapi2.init: setDebugVerbosity")
		if (type(commonapi._info.settings.verbosity) == "number") then
			commonapi._native.setDebugVerbosity(commonapi._info.settings.verbosity)
		end
		--print("commonapi2.init: setSettings")
		commonapi._native.setSettings(commonapi._info.settings)
		--print("commonapi2.init: doPatches")
		commonapi._native.doPatches()


		--print("commonapi2.init: loadTranslationSystem", commonapi._native.getLanguage() )

		if ( commonapi._native.getGUIBindings ~= nil) then
			--commonapi.ui = commonapi._nativeui
			commonapi.ui._activate()
		end

		if (type(commonapi._info.settings.uiscale) == "number") then
			--commonapi._native.setUIScale(commonapi._info.settings.uiscale)
		end

		if (commonapi._events ~= nil) then
			if (commonapi._events._activate()) then
				commonapi.events = commonapi._events
			end
		end

--[[		if (commonapi._vehanimations ~= nil) then
			if (commonapi._vehanimations._activate()) then
				commonapi.vehanimations = commonapi._vehanimations
			end
		end
--]]


--[[
		commonapi_internal._unique_threadid = commonapi._native.getLuaStatePtr()
		commonapi_internal._unique_threadid_check = function()
			local ourthreadid = commonapi._native.getLuaStatePtr()
			if (commonapi_internal._unique_threadid == ourthreadid)  then
				print("commonapi2._unique_threadid_check(n) threadid match = "  ..ourthreadid)
			else
				print("commonapi2._unique_threadid_check(native) MISMATCH ".. commonapi_internal._unique_threadid .. " !="..ourthreadid)
			end
		end
--]]
	end
end



local function loadTranslationSystem(translations)
	if (setStrings == nil) then
		require "mod"
	end
	setStrings("eis_os_commonapi2_1", translations)

	if (getTextRes == nil or getTextRes == false) then
		getTextRes = function(s)
			return s
		end
	end

	local lang = "en"

	if (commonapi._native ~= nil) then
		lang = commonapi._native.getLanguage()
	end

	if (_ == nil) then
		local translatefn = 'function _(s) return translateModStr("eis_os_commonapi2_1", "'..lang..'", s) end'
		load(translatefn)()
	end

	if (commonapi._native ~= nil) then
		if (translations[lang] ~= nil) then
			commonapi._native.setTranslations(translations[lang])
		else
			commonapi._native.setTranslations(translations["en"])
		end
	end
end




local function renameUGConsoleState()
	if (app ~= nil) then
		if ( type(app) == "table" and app.loadGame) then
			if (commonapi._native ~= nil and commonapi._native.renameLuaState ~= nil) then
				commonapi._native.renameLuaState("UG Console State")
			end
		end
	end
end


local function detectUGAPIChanges()
	local mtapichecker = {}
	local mtapiguiutilchecker = {}
	local mtapiguiutilgetByIdchecker = {}
	mtapiguiutilgetByIdchecker.__newindex = function(t,k,v)
		rawset(t, k, v)
		if (k == "__doc__") then
			local failback = require("commonapi2.ui2.failback")
			failback.init()
		end
	end


	mtapiguiutilchecker.__newindex = function(t,k,v)
		rawset(t, k, v)
		if (k == "util" and type(v) == "table") then
			setmetatable(t[k], {
				__newindex = function(tx,kx,vx)
					rawset(tx, kx, vx)
					if (kx == "getById") then
						setmetatable(vx, mtapiguiutilgetByIdchecker)
					end
				end
			})
		end
	end

	mtapichecker.__newindex = function(t,k,v)
		rawset(t, k, v)
		if (k == "gui" and type(v) == "table") then
			if (app ~= nil) then
				if ( type(app) == "table" and app.loadGame) then
					setmetatable(v, mtapiguiutilchecker)
					renameUGConsoleState()
				end
			end
		end
	end
	if (api ~= nil) then
		setmetatable(api, mtapichecker)
	end
end

function commonapi._init(mystrings)
	if (_G['commonapi'] ~= nil) then
		local otherversion = _G['commonapi']
		if (otherversion.version ~= commonapi.version) then
			local twoversions = "\n\nERROR: CommonAPI2 with different version installed! You can only have one CommonAPI2 installed!\n\n"
			error("VERSIONS MISMACH(_init)" ..twoversions, 2)
		end
-- CommonAPI strings.lua could be loaded again in the same context
--		print("ERROR: Common API was loaded already? You can only have one commonapi installed!")
		return false
	end
	print("commonapi2.init", commonapi.version)
	_G['commonapi'] = commonapi

	detectOS()
	loadSettings()
	loadSettingsDefaults()

	if (commonapi._info.settings.disablenative == true) then
		print("commonapi.init: Native code loading disabled by settings.lua")
	else
		loadCommonAPILuaExtension(mystrings)
	end
	--print("commonapi.init: loadTranslationSystem")
	loadTranslationSystem(mystrings)

	-- Need to load early for proper hooking...
	local mods = commonapi.mods

	detectUGAPIChanges()
	return true
end


return commonapi