--- Nodedata interface module
-- @author Oskar Eisemuth
-- @copyright 2023
-- @module commonapi.nodedata
local nodedata = {}

local commonapimath = require("commonapi2._math")
local svec3 = commonapimath.svec3
local vec3 = commonapimath.vec3

local function mathround(num, deci)
	local d = 10 ^ (deci or 2)
	num = num * d
	if num >= 0 then
		num = math.floor(num + 0.5)
	else
		num = math.ceil(num - 0.5)
	end
	return num / d
end


function nodedata.getNodesData()
	return commonapi._native.FlexStreets.getNodesData()
end

function nodedata.setNodeData(data)
	--print("NodeWindow:setNodeData(nodeData)")
	--commonapi.dmp(nodeData)
	commonapi._native.FlexStreets.setNodeDataById(data.id, data)

	--local nodeData2 = commonapi._native.FlexStreets.getNodeDataById(nodeData.id)
	--commonapi.dmp(nodeData2)
	--print("NodeWindow:setNodeData(nodeData) end")
end

function nodedata.deleteNodeData(data)
	if (data.id ~= nil) then
		commonapi._native.FlexStreets.setNodeDataById(data.id, nil)
	end
end

function nodedata.getNodeDataById(id)
	local data
	if (id == nil) then
		data = commonapi._native.FlexStreets.createNodeData()
	else
		data = commonapi._native.FlexStreets.getNodeDataById(id)
		if (data ~= nil) then
			data.id = id
		end
	end
	return data
end

function nodedata.getNodesDataByPosition(pos, margin)
	return commonapi._native.FlexStreets.getNodesDataByPos(pos[1], pos[2], margin)
end

function nodedata.getOrCreateNodeDataForPosition(pos)
	if (pos[3] == nil) then
		pos[3] = 0
	end
	local radius = 2.0
	local nodesData = nodedata.getNodesDataByPosition(pos, radius)

	--print("getOrCreateNodeDataForPosition #nodesData = "..#nodesData, pos[1], pos[2], radius )
	local firstNodeIndex, firstNodeData = next(nodesData)
	if (firstNodeIndex) then
		return firstNodeData
	end

	local entities = game.interface.getEntities({
		radius = 2.0,
		pos = {
			pos[1],
			pos[2],
			pos[3],
		}
	}, { includeData = true, type = "BASE_NODE" })

	local entityselected = nil
	local entityselectedoffset = nil

	for entityid, entityData in pairs(entities) do
		local diff = svec3.lengthBetween(pos, entityData.position)

		if (entityselectedoffset == nil or (diff < entityselectedoffset)) then
			--print("previous entityselectedoffset: ", entityselectedoffset, " diff:", diff, " new entity id:", entityid)
			entityselectedoffset = diff
			entityselected = entityid
		end
	end

	if (entityselected ~= nil) then
		local entityData = entities[entityselected]
		-- Try again with BASE_NODE location
		local entityDataPos = { entityData.position[1], entityData.position[2], entityData.position[3] }
		nodesData = nodedata.getNodesDataByPosition(entityDataPos, radius)
		firstNodeIndex, firstNodeData = next(nodesData)
		if (firstNodeIndex) then
			return firstNodeData
		end

		local nodeData = commonapi._native.FlexStreets.createNodeData()
		if (nodeData.id == nil or nodeData.id == "") then
			math.randomseed(os.time())
			nodeData.id = "ND"..mathround(entityData.position[1], 0).."_"..mathround(entityData.position[2], 0).."_"..math.random(0, 100)
			--nodeData.id = "ND"..math.random(0, 100000)
		end
		nodeData.pos = { entityData.position[1],  entityData.position[2], entityData.position[3] }
		return nodeData
	end
	return nil
end



function nodedata.rebuildNode(pos, callback, ignoreCollision)
	local entitiesBaseNode = game.interface.getEntities({
		radius = 0.3,
		pos = {
			pos[1],
			pos[2],
			0
		}
	}, { includeData = true, type = "BASE_NODE" })

	local baseNodeEntityId = nil
	local firstEntityIndex, entityData = next(entitiesBaseNode)
	if (firstEntityIndex) then
		baseNodeEntityId = entityData.id
	end

	if (baseNodeEntityId == nil) then
		return
	end


	local n2street = api.engine.system.streetSystem.getNode2StreetEdgeMap()
	local connectedStreets = n2street[baseNodeEntityId] or {}
	local proposal = api.type.SimpleProposal.new()
	local idx = 1
	local entities = {}
	local conEntities = {}

	for edgeEntityIdx, edgeEntityId in ipairs(connectedStreets) do

		local conEntity = api.engine.system.streetConnectorSystem.getConstructionEntityForEdge(edgeEntityId)
		if ( conEntity > -1 ) then
			-- pass
		else
			entities[#entities+1] = edgeEntityId
			local baseEdge = api.engine.getComponent(edgeEntityId, api.type.ComponentType.BASE_EDGE)
			local baseEdgeStreet = api.engine.getComponent(edgeEntityId, api.type.ComponentType.BASE_EDGE_STREET)
			local baseEdgeTrack = api.engine.getComponent(edgeEntityId, api.type.ComponentType.BASE_EDGE_TRACK)
			local eo = api.type.SegmentAndEntity.new()
			eo.entity = -idx
			eo.type = 0
			eo.comp = baseEdge
			if (baseEdgeStreet ~= nil) then
				local playerOwned = api.engine.getComponent(edgeEntityId, api.type.ComponentType.PLAYER_OWNED)
				if (playerOwned ~= nil) then
					local playerOwnedComponent = api.type.PlayerOwned.new()
					playerOwnedComponent.player = playerOwned.player
					eo.playerOwned = playerOwnedComponent
				end
				eo.streetEdge = baseEdgeStreet
				eo.type = 0
			end
			if (baseEdgeTrack ~= nil) then
				eo.type = 1
				eo.trackEdge = baseEdgeTrack
			end
			proposal.streetProposal.edgesToRemove[idx] = edgeEntityId
			proposal.streetProposal.edgesToAdd[idx] = eo
			idx = idx + 1
		end
	end

	local ctx2 = api.type.Context:new()
	ctx2.checkTerrainAlignment = false
	ctx2.cleanupStreetGraph = false
	ctx2.gatherBuildings = true
	ctx2.gatherFields = true

	local allowCollision = (ignoreCollision == true) or false
	local cmd = api.cmd.make.buildProposal(proposal, ctx2, allowCollision)

	local fn = function(result, success)
		if (result ~= nil and result.resultProposalData ~= nil and result.resultProposalData.errorState) then
			local errorState = result.resultProposalData.errorState
			if (#errorState.messages > 0) then
				local messages = {}
				for j, msg in ipairs(errorState.messages) do
					messages[#messages+1] = msg
				end
				local msg = table.concat(messages, "\n")
				pcall(callback, success, msg)
				return
			end
		end
		pcall(callback, success)
	end
	api.cmd.sendCommand(cmd, fn)
end

--commonapi.nodedata.removeAll()
function nodedata.removeAll()
	local nodes = nodedata.getNodesData()
	local count = 0
	for _ in pairs(nodes) do
		count = count + 1
	end
	print("nodedata.removeAll: trying to rebuild nodecount: "..count)
	io.flush()
	commonapi._native.FlexStreets.clearNodesData()
	print("nodedata.removeAll: trying to rebuild nodecount: "..count)
	io.flush()
	for i, node in pairs(nodes) do
		local fn = function(success, message)
			if (success ~= true) then
				print("node.pos: "..node.pos[1].. ", "..node.pos[2].."): "..message)
			end
		end
		nodedata.rebuildNode(node.pos, fn, true)
	end
	print("nodedata.removeAll: commands to rebuild nodes send..")
	io.flush()
end

--[[
	idx = 0; api.engine.forEachEntityWithComponent(function(id) idx=idx+1 end, api.type.ComponentType.BASE_NODE); print("Visited BASE_NODEs: "..idx)
commonapi.nodedata.runMapTest()
commonapi.nodedata.reload()

pos = { 668, -6442 }
commonapi.nodedata.getNodesDataByPosition(pos, 20)
--]]
function nodedata.runMapTest()
	local testtimer = os.clock()
	local idx = 0
	local fn = function(id)
		idx = idx + 1
		local c = api.engine.getComponent(id, api.type.ComponentType.BASE_NODE)
		local data = nodedata.getOrCreateNodeDataForPosition( {c.position.x, c.position.y } )
		if (data.id == nil or data.id == "") then
			print("Error in getOrCreateNodeDataForPosition for", idx)
			commonapi.dmp(data)
		else
			nodedata.setNodeData(data)
		end
	end
	api.engine.forEachEntityWithComponent(fn, api.type.ComponentType.BASE_NODE)

	print(string.format("elapsed time for creating nodes: %.2f", os.clock() - testtimer))


	print("Visited BASE_NODEs: ", idx)
	--local all = nodedata.getNodesData()
	--print("Currently nodes in memory: ", #all)

	testtimer = os.clock()
	local pos = { 668, -6442, 0 }
	for i = 0, 10000 do
		local nodesData = nodedata.getNodesDataByPosition(pos, 20)
	end
	local nodesDataX = nodedata.getNodesDataByPosition(pos, 20)
	commonapi.dmp(nodesDataX)
	print(string.format("elapsed time for getNodesDataByPosition: %.2f", os.clock() - testtimer))
end

function nodedata.reload()
	commonapi._reloadModule("nodedata")
end
return nodedata