local ts = {}
local header = "function data()\nreturn"
local footer = "\nend\n"

local function writeKey(ctx, k)
	if type(k) == "string" and string.find(k, "^[_%a][_%w]*$") then
		ctx.write(k.." = ")
	else
		ctx.write("[")
		ctx:s(k, "")
		ctx.write("] = ")
	end
end

local fns = {}
fns["unknown"] = function(self, o)
-- do nothing
end
fns["string"] = function(self, o) self.write(string.format("%q", o)) end
fns["nil"] = function(self, o) self.write("nil") end
fns["boolean"] = function(self, o) self.write(tostring(o)) end
fns["number"] = function(self, o) self.write(o) end
fns["function"] = function(self, o) self.write("<function>") end
fns["table"] = function(self, o, prefix)
		local metatag = o["__metatag__"]
		if metatag then
			if metatag == 0 then
				self.write("_(")
				self:s(o.val, prefix)
				self.write(")")
			else
				error("invalid metatag: " .. metatag)
			end
			return
		end

		local oneLine = true
		local listKeys = {}
		local tableKeys = {}
		for k,v in ipairs(o) do
			listKeys[k] = true
		end
		for k,v in pairs(o) do
			if type(v) == "table" then oneLine = false end
			if not listKeys[k] then
				tableKeys[#tableKeys+1] = k
				oneLine = false
			end
		end
		table.sort(tableKeys, function(a, b) return type(a) < type(b) or (type(a) == type(b) and a < b) end)

		local prefix2 = prefix .. self.indentSym
		local maxCount = self.maxCount
		local count = 0

		if oneLine then
			self.write("{ ")
			for k,v in ipairs(o) do
				count = count + 1
				if count >= maxCount then self.write("[truncated] ") break end
				--self:s(self, v, "") -- make tables fast again:
				local fn = fns[type(v)] or fns.unknown
				fn(self, v, "")
				self.write(", ")
			end
			for i,k in ipairs(tableKeys) do
				count = count + 1
				if count >= maxCount then self.write("[truncated] ") break end

				local v = o[k]
				self:writeKey(k)
				--self:s(v, "")	-- make tables fast again:
				local fn = fns[type(v)] or fns.unknown
				fn(self, v, "")
				self.write(", ")
			end
			self.write("}")
		else
			self.write("{\n")
			for k,v in ipairs(o) do
				count = count + 1
				if count >= maxCount then self.write("[truncated] ") break end

				self.write(prefix2)
				--self:s(v, prefix2)	-- make tables fast again:
				local fn = fns[type(v)] or fns.unknown
				fn(self, v, prefix2)
				self.write(",\n")
			end
			for i,k in ipairs(tableKeys) do
				count = count + 1
				if count >= maxCount then self.write("[truncated] ") break end

				local v = o[k]
				self.write(prefix2)
				self:writeKey(k)
				--self:s(v, prefix2)	-- make tables fast again:
				local fn = fns[type(v)] or fns.unknown
				fn(self, v, prefix2)
				self.write(",\n")
			end
			self.write(prefix)
			self.write("}")
		end
	end

fns["userdata"] = function(self, o, prefix)
		local mt = getmetatable(o)
		local okpairs, vpr = pcall(function() return mt.pairs end)
		local okmem, members = pcall(function() return mt.__members end)
		local prefix2 = prefix .. self.indentSym
		local maxCount = self.maxCount
		local count = 0

		if mt and okpairs and vpr then
			self.write("{\n")
			for k, v in pairs(o) do
				count = count + 1
				if count >= maxCount then
					self.write("[truncated] ")
					break
				end

				self.write(prefix2)
				self:writeKey(k)
				self:s(v, prefix2)
				self.write(",\n")
			end
			self.write(prefix)
			self.write("}")
		elseif mt and okmem and members then
			self.write("{\n")
			for i = 1, #members do
				count = count + 1
				if count >= maxCount then
					self.write("[truncated] ")
					break
				end

				local k = members[i]
				local l, v = pcall(function() return o[k] end)
				if l then
					self.write(prefix2)
					self:writeKey(k)
					self:s(v, prefix2)
					self.write(",\n")
				end
			end
			self.write(prefix)
			self.write("}")
		else
			self.write(tostring(o))
		end
	end


local function ctxSerialize(ctx, o, prefix)
	local fn = fns[type(o)] or fns.unknown
	fn(ctx, o, prefix)
end

local function createCtx(writeFn, indentSym, maxCount)
	local ctx = {
		maxCount = maxCount or 1073741824,
		write = writeFn,
		indentSym = indentSym or "\t",
		s = ctxSerialize,
		writeKey = writeKey,
	}
	return ctx
end


function ts.serialize(o, maxCount)
	io.write(header)
	local ctx = createCtx(io.write, nil, maxCount)
	ctx:s(o, "")
	io.write(footer)
	io.flush()
end

function ts.serializeStr(o, maxCount)
	local buf = {
		header
	}
	local function writeFn(v)
		buf[#buf+1] = ""..v
	end

	local ctx = createCtx(writeFn, nil, maxCount)
	ctx:s(o, "")
	buf[#buf+1] = footer
	return table.concat(buf)
end

function ts.toString(o, sep, maxCount)
	local buf = {}

	local function writeFn(v)
		buf[#buf+1] = ""..v
	end
	local ctx = createCtx(writeFn, "  ", maxCount)
	ctx:s(o, "")
	return table.concat(buf)
end

return ts