local NodeWindow = {}

local ui = require("commonapi2.ui2.tpfui")
local nd = commonapi.nodedata

local zoneutil = require "mission.zone"
local commonapimath = require("commonapi2._math")
local svec3 = commonapimath.svec3
local svec2 = commonapimath.svec2
local vec3 = commonapimath.vec3

local secondaryui = require "commonapi2.ui2.secondaryui"

require("serialize")
NodeWindow = ui.Window:new()

NodeWindow.id = "NodeWindow"
NodeWindow.caption = "Node editor"
NodeWindow.uilist = {}

NodeWindow.height = 500
NodeWindow.width = 500
NodeWindow._zoneid = "commonapi2_NodeWindow_zone"


local function mathround(num, deci)
	local d = 10 ^ (deci or 2)
	num = num * d
	if num >= 0 then
		num = math.floor(num + 0.5)
	else
		num = math.ceil(num - 0.5)
	end
	return num / d
end

function NodeWindow.create()
	local o
	o = NodeWindow:new()
	o:createUIList()
	o:open()
	o:setPositionCenter()
	return o
end

function NodeWindow:onclose()
	self:clearNodeMarkers()
	self:clearCircleMarkers()
	self:installShowMapMarkersCB(false)
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end
	local curRenderSettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
	curRenderSettings["showLanes"] = false
end

function NodeWindow:refresh()
	self:createUIList()
	self:rebuild()
end


function NodeWindow:createUIList()
	self.uilist = self:createUIListTabNodeEditor()
end


function NodeWindow:addNodeMarker(id, pos, color, radius, parts)
	if (self._nodemarkers == nil) then
		self._nodemarkers = {}
	end
	if (color == nil) then
		color = {.8, .8, .8, .7}
	end
	if (radius == nil) then
		radius = 30
	else
		radius = radius * 25
	end
	local node = {
		pos = pos,
		color = color,
		radius = radius,
	}
	self._nodemarkers[id] = node
	commonapi._native.FlexStreets.setTNRPoints( self._nodemarkers )
end

function NodeWindow:clearNodeMarkers()
	commonapi._native.FlexStreets.setTNRPoints()
	self._nodemarkers = nil
end


function NodeWindow:addCircleMarker(id, pos, color, radius, parts)
	if (self._circlemarkers == nil) then
		self._circlemarkers = {}
	end
	if (color == nil) then
		color = {.8, .8, .8, .7}
	end
	if (radius == nil) then
		radius = 10
	end
	if (parts == nil) then
		parts = 9
	end
	local zonedata = {
		polygon = zoneutil.makeCircleZone(pos, radius, parts),
		draw = true,
		drawColor = color
	}
	self._circlemarkers[id] = true
	game.interface.setZone(id, zonedata)
end

function NodeWindow:clearCircleMarkers()
	if (self._circlemarkers == nil) then
		return
	end
	for id, v in pairs(self._circlemarkers) do
		game.interface.setZone(id)
	end
	self._circlemarkers = nil
end

function NodeWindow:installSelectNode()
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end

	local win = self
	secondaryui.setMouseListener({
		filterui = false,
		fn = function(evt, mouseScreenPos)
			--return win:onFindPolygonMouseEvent(evt, mouseScreenPos)
			local ok, result = pcall(win.onFindBaseNodeMouseEvent, win, evt, mouseScreenPos)
			if (ok == false) then
				print(result)
				return false
			end
			return result
		end,
		cleanup = function()
			local ok, err = pcall(win.onFindBaseNodeMouseEvent_cleanup, win)
			if (ok == false) then
				print(err)
			end
		end,
	})
end


function NodeWindow:onFindBaseNodeMouseEvent_cleanup()
	self:refresh()
end

function NodeWindow:onFindBaseNodeMouseEvent(evt, mouseScreenPos)
	if (self._ugwindow == nil) then
		print("NodeWindow:onFindBaseNodeMouseEvent lost self._ugwindow")
		secondaryui.setMouseListener(nil)
		return
	end
	if ( (evt.type == 2 and evt.button == 1) ) then
		--print("NodeWindow:onFindBaseNodeMouseEvent evt.button == 1 found")
		secondaryui.setMouseListener(nil)
		self:refresh()
		return
	end

	local _ugwindow = self._ugwindow
	local rect = _ugwindow:getContentRect()
	if (rect:contains(mouseScreenPos.x, mouseScreenPos.y)) then
		return
	end

	if (evt.type == 2 and evt.button == 2) then
		local pos = game.gui.getTerrainPos()
		if (pos == nil) then
			return
		end
		self:clickRightOnMap()
		return true
	end

	if (evt.type == 2 and evt.button == 0) then
		local pos = game.gui.getTerrainPos()
		if (pos == nil) then
			return
		end
		self:clickLeftOnMap(pos)
		return true
	end
end



function NodeWindow:selectState(stateName, data)
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end
	if (stateName == nil) then
		self.curstate = nil
		self:clearNodeMarkers()
		self:clearCircleMarkers()
		self:installShowMapMarkersCB(false)
		self:refresh()
		return
	end
	local o = data or {}

	o.name = stateName
	local mt = {
		__index = NodeWindow[stateName],
	}
	setmetatable(o, mt)

	self.curstate = o
	if (self.curstate ~= nil and self.curstate.init ~= nil) then
		local ok, err = pcall(self.curstate.init, self, self.curstate)
		if (ok == false) then
			print(err)
		end
	end
	self:installSelectNode()
	self:clearNodeMarkers()
	self:clearCircleMarkers()
	self:installShowMapMarkersCB((o.showMapMarkers ~= nil))
	self:refresh()
end


function NodeWindow:showEditorPane(uilist)
	if (self.curstate ~= nil and self.curstate.showEditorPane ~= nil) then
		return self.curstate.showEditorPane(self, self.curstate, uilist)
	end
	return false
end

function NodeWindow:clickLeftOnMap(pos)
	if (self.curstate ~= nil) then
		if (self.curstate.clickLeftOnMap ~= nil) then
			return self.curstate.clickLeftOnMap(self, self.curstate, pos)
		end
	end
end

function NodeWindow:clickRightOnMap(pos)
	if (self.curstate ~= nil) then
		if (self.curstate.clickRightOnMap ~= nil) then
			return self.curstate.clickRightOnMap(self, self.curstate, pos)
		end
	end
end

function NodeWindow:showMapMarkers(cameraPos)
	if (self.curstate ~= nil) then
		if (self.curstate.showMapMarkers ~= nil) then
			return self.curstate.showMapMarkers(self, self.curstate, cameraPos)
		end
	end
	return false
end



--- Select Node
NodeWindow.stateSelectNode = {}

function NodeWindow.stateSelectNode:showEditorPane(statedata, uilist)
	uilist[#uilist + 1] = {
		typ = "Label",
		caption = self._("Select a red or blue base node by mouse left click,"),
		styleclass = "commonapi-nodeeditor-red",
	}
	uilist[#uilist + 1] = {
		typ = "Label",
		caption = self._("right click to leave tool mode"),
		styleclass = "commonapi-nodeeditor-colorpadding"
	}
	return uilist
end

function NodeWindow.stateSelectNode:clickRightOnMap(statedata, pos)
	self:selectState(nil)
end

function NodeWindow.stateSelectNode:clickLeftOnMap(statedata, pos)
	local nodeData = nd.getOrCreateNodeDataForPosition(pos)
	if (nodeData ~= nil) then
		self:selectState("stateNodeEditor", {
			nodeDataId = nodeData.id,
			nodeData = nodeData,
			advMode = false,
		})
	end
end

function NodeWindow.stateSelectNode:showMapMarkers(statedata, cameraPos)
	local radius = 300
	local curpos = { cameraPos[1], cameraPos[2], 0 }
	local nodesData = nd.getNodesDataByPosition(curpos, radius)
	for k, nodeData in ipairs(nodesData) do
		--if (statedata.nodeDataId ~= nodeData.id) then
			self:addCircleMarker(nodeData.id, nodeData.pos, { 0.0, 0.0, 1.0, 0.7}, 11)
	--	else
	--		self:addCircleMarker(nodeData.id, nodeData.pos, { 0.0, 1.0, 0.0, 0.7}, 11)
	--	end
	end

	local entities = game.interface.getEntities({
		radius = radius,
		pos = {
			cameraPos[1],
			cameraPos[2],
			cameraPos[3]
		}
	}, { includeData = true, type = "BASE_NODE" })

	for k, entityData in pairs(entities) do
		if (entityData.position ~= nil) then
			self:addCircleMarker("entity"..k, { entityData.position[1],  entityData.position[2] }, { 1.0, 0.0, 0.0, 0.7})
		end
	end
end



--- Edit selected Node
NodeWindow.stateNodeEditor = {}

local function table2int(t)
	local result = 0
	for i = 1, #t do
		if (t[i] == 1) then
			result = result + (2 ^ (i-1))
		end
	end
	return result
end


function NodeWindow.stateNodeEditor:init(statedata)
	if (statedata.nodeData == nil) then
		return
	end

	local curRenderSettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
	curRenderSettings["showLanes"] = true

	local pos = statedata.nodeData.pos

	local entitiesBaseNode = game.interface.getEntities({
		radius = 0.3,
		pos = {
			pos[1],
			pos[2],
			0
		}
	}, { includeData = true, type = "BASE_NODE" })

	local baseNodeEntityId = nil
	local firstEntityIndex, entityData = next(entitiesBaseNode)
	if (firstEntityIndex) then
		baseNodeEntityId = entityData.id
	end

	if (baseNodeEntityId == nil) then
		return
	end
	statedata.baseNodeEntityId = baseNodeEntityId

	statedata.baseNodePos = entityData.position

	statedata.baseNodeRadius = 3;


	local n2street = api.engine.system.streetSystem.getNode2StreetEdgeMap()
	local connectedStreets = n2street[baseNodeEntityId] or {}


	local rad90 = math.rad(90)
	--api.engine.getComponent(statedata.baseNodeEntityId, api.type.ComponentType.SHAPE_LIST)
	--baseNodeEntityId
	local streetData = {}
	for idx, edgeEntityId in ipairs(connectedStreets) do

		local compBaseEdge = api.engine.getComponent(edgeEntityId, api.type.ComponentType.BASE_EDGE)
		local compBaseEdgeStreet =  api.engine.getComponent(edgeEntityId, api.type.ComponentType.BASE_EDGE_STREET)
		local compShapeList = api.engine.getComponent(edgeEntityId, api.type.ComponentType.SHAPE_LIST)

		local shape = compShapeList.shapes[1]
		local attributes = commonapi._native.StreetTypes.getShapeAttributes(shape)
		local attrPoints = attributes.data[13]

		local tangent
		local xpos
		local isReversed = false
		if (compBaseEdge.node0 == baseNodeEntityId) then
			--tangent = svec3.normalize( { compBaseEdge.tangent0.x, compBaseEdge.tangent0.y, compBaseEdge.tangent0.z } )
			local attrPoint = attrPoints[1]
			local xtangent = attrPoints[3]
			tangent =  svec3.normalize( { xtangent.x, xtangent.y, xtangent.z } )
			xpos = { attrPoint.x, attrPoint.y, attrPoint.z }
		end
		if (compBaseEdge.node1 == baseNodeEntityId) then
			--tangent = svec3.normalize( { -compBaseEdge.tangent1.x, -compBaseEdge.tangent1.y, -compBaseEdge.tangent1.z } )
			local attrPoint = attrPoints[2]
			local xtangent = attrPoints[4]
			tangent =  svec3.normalize( { -xtangent.x, -xtangent.y, -xtangent.z } )
			xpos = { attrPoint.x, attrPoint.y, attrPoint.z }
			isReversed = true
		end

		local length2xpos = svec3.lengthBetween(xpos, pos)
		if (statedata.baseNodeRadius < length2xpos) then
			statedata.baseNodeRadius = length2xpos
		end


		if (tangent ~= nil and compBaseEdgeStreet ~= nil) then
			local angleRad = math.atan2(tangent[2], tangent[1])
			local angle = math.deg(angleRad)
			if (angle < 0) then
				angle = angle + 360
			end
			--local xpos =  svec3.add(statedata.baseNodePos,  svec3.mul(5, tangent) )

			local laneAnchors = {}


			local trotate = svec3.normalize(svec3.rotateZRad(tangent, -rad90))
			local streetType = api.res.streetTypeRep.get(compBaseEdgeStreet.streetType)
			local laneConfigs = streetType.laneConfigs

			local edgeHalfWidth = streetType.streetWidth/2
			local startPoint = svec3.add(xpos, svec3.mul(-edgeHalfWidth, trotate))

			local posOrth = 0



			if (isReversed == true) then
				local laneConfigsSize = #laneConfigs
				for i = 2, laneConfigsSize-1, 1 do
					local laneConfig = laneConfigs[ (laneConfigsSize+1-i) ]
					local startToLaneNode = svec3.mul( (posOrth + laneConfig.width/2), trotate)
					local laneNodePos = svec3.add(startPoint, startToLaneNode)
					posOrth = posOrth + laneConfig.width
					local anchorRadius = laneConfig.width/2
					if (anchorRadius < 0.5) then
						anchorRadius = 0.5
					end
					laneAnchors[#laneAnchors+1] = {
						idx = i,
						pos = laneNodePos,
						radius = anchorRadius,
						transportModes = table2int(laneConfig.transportModes),
						forward = (laneConfig.forward == false)
					}
				end
			else
				for i = 2, #laneConfigs-1, 1 do
					local laneConfig = laneConfigs[i]
					local startToLaneNode = svec3.mul( (posOrth + laneConfig.width/2), trotate)
					local laneNodePos = svec3.add(startPoint, startToLaneNode)
					posOrth = posOrth + laneConfig.width
					local anchorRadius = laneConfig.width/2
					if (anchorRadius < 0.5) then
						anchorRadius = 0.5
					end
					laneAnchors[#laneAnchors+1] = {
						idx = i, --#laneConfigs+1-i,
						pos = laneNodePos,
						radius = anchorRadius,
						transportModes = table2int(laneConfig.transportModes),
						forward = (laneConfig.forward == true)
					}
				end
			end

			streetData[#streetData+1] = {
				entityId = edgeEntityId,
				streetTypeId = compBaseEdgeStreet.streetType,
				tangent = tangent,
				angle = angle,
				angleRad = angleRad,
				xpos = xpos,
				laneAnchors = laneAnchors,
				isReversed = isReversed
			}
		end
	end
	statedata.streetData = streetData
	statedata.connectionMap = {}

	-- internal to connection map
	statedata.loadConnecttionMap(self, statedata)
end


function NodeWindow.stateNodeEditor:showEditorPane(statedata, uilist)
	uilist[#uilist + 1] = {
		typ = "Label",
		caption = self._("Please select a source lane node by left clicking on highlighted green anchor,"),
		styleclass = "commonapi-nodeeditor-green"
	}

	uilist[#uilist + 1] = {
		typ = "Label",
		caption = self._("right click to exit and back to node selection."),
		styleclass = "commonapi-nodeeditor-colorpadding"
	}

	uilist[#uilist + 1] = {
		typ = "Button",
		id = "selectedNodeData",
		data = {
			nodeDataId = statedata.nodeDataId,
		},
		caption = self._("Currently selected node data:").." " .. statedata.nodeDataId,
		onclick = function(item, win)
			commonapi._native.utils.setClipboardText("commonapi._native.FlexStreets.getNodeDataById(\""..item.data.nodeDataId.."\")")
		end
	}

	uilist[#uilist + 1] = {
		typ = "Button",
		id = "deleteNodeData",
		caption = self._("Delete additional data and reset base node"),
		styleclass = "color-button",
		onclick = function(item, win)
			win:deleteNodeDataButton()
		end
	}

	if (self.params.showDebugData == true) then

	local uiTableData = {}
	local columnHeaders = {
		{
			caption = "Idx",
			width = 10
		},
		{
			caption = "Street Entity Id",
			width = 50
		},
		{
			caption = _("Type, LaneId"),
			width = 3
		},
		{
			caption = _("Position"),
			width = 50
		},
		{
			caption = _(""),
			width = 50
		},
		{
			caption = _("Tangent / Lane info"),
			width = 50
		},
		{
			caption = self._("Angle"),
			width = 50
		},
	}

	local anchormapping = {}

	uilist[#uilist + 1] = {
		typ = "Table",
		id = "StreetsToNodeTable",
		columnHeaders = columnHeaders,
		columns = #columnHeaders,
		border = false,
		uilist = uiTableData,
		size = {
			0,
			-100
		},
		flags = {
			Sortable = false,
			MultiSortable = false,
			ScrollFreezeTopRow = true
		},
		onclick = function(item, win)
			local index = win.params["StreetsToNodeTable"]
			if (index >= 0 and anchormapping) then
				local data = anchormapping[index + 1]
				if (data) then
					if (data.laneAnchor) then
						self:addCircleMarker("StreetsToNodeTable", data.laneAnchor.pos, { 1.0, 0.0, 0.1, 1.0}, 2, 7)
					end
				end
			end
		end
	}
	local function createItem(streetIdx, streetData)
		anchormapping[#anchormapping+1] = {
			streetIdx = streetIdx
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..streetIdx,
			column = 0,
			rownew = true,
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..streetData.entityId,
			column = 1,
		}
		local isReversed = ""
		if (streetData.isReversed) then
			isReversed = " (reversed)"
		end
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..streetData.streetTypeId ..isReversed,
			column = 2
		}

		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..mathround(streetData.xpos[1], 2),
			column = 3
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..mathround(streetData.xpos[2], 2),
			column = 4
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..mathround(streetData.tangent[1], 2) .. ", "..mathround(streetData.tangent[2], 2),
			column = 5
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..mathround(streetData.angle, 2) .. "°",
			column = 6
		}


		for laneIdx, laneAnchor in pairs(streetData.laneAnchors) do
			anchormapping[#anchormapping+1] = {
				streetIdx = streetIdx,
				laneAnchor = laneAnchor
			}
			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = "",
				column = 0,
				rownew = true,
			}

			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = "",
				column = 1,
			}

			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = ""..laneAnchor.idx,
				column = 2
			}

			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = ""..mathround(laneAnchor.pos[1], 2),
				column = 3
			}
			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = ""..mathround(laneAnchor.pos[2], 2),
				column = 4
			}
			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = "forward = "..tostring(laneAnchor.forward == true).." transportModes =  "..tostring(laneAnchor.transportModes),
				column = 5
			}

			uiTableData[#uiTableData + 1] = {
				typ = "Label",
				caption = "",
				column = 6
			}
		end
	end



	if (statedata.streetData ~= nil) then
		for id, streetData in pairs(statedata.streetData) do
			createItem(id, streetData)
		end
	end

	end -- showDebugData
	return uilist
end
--]]

function NodeWindow.stateNodeEditor:clickRightOnMap(statedata, pos)
	self:selectState("stateSelectNode", {})
end

function NodeWindow.stateNodeEditor:clickLeftOnMap(statedata, pos)
	if (statedata.streetData ~= nil) then
		local filterForward
		if (statedata.advMode ~= true) then
			filterForward = false
		end
		local streetDataIdx, laneAnchorIdx, laneAnchor = statedata:getStateLaneAnchorByPos(pos, filterForward)
		if (streetDataIdx ~= nil and (laneAnchor.forward == false or statedata.advMode == true)) then
			self:selectState("stateSelectLaneAnchor", {
				editor = statedata,
				srcStreetDataIdx = streetDataIdx,
				srcLaneAnchorIdx = laneAnchorIdx
			})
		end
	end
end


local function setTableValue(t, keys, value)
	local kc = #keys
	for i = 1, kc - 1 do
		local k = keys[i]
		if (t[k] == nil) then
			t[k] = {}
		end
		t = t[k]
	end
	t[keys[kc]] = value
end

function NodeWindow.stateNodeEditor:loadConnecttionMap(statedata)
	local addCons = statedata.nodeData.addCons or {}
	local removeCons = statedata.nodeData.removeCons or {}
	local streetIdxToAngle = statedata.nodeData.streetIdxToAngle or {}

	local streetIdxToIdx = {}
	local streetIdxToIdxAngle = {}
	local connectionMap = {}
	--commonapi.dmp(removeCons)
	statedata.connectionMap = connectionMap
	for streetIdxAngle, angle in ipairs(streetIdxToAngle) do
		local found = false
		for streetIdx, streetData in pairs(statedata.streetData) do
			local diff = math.abs(streetData.angle - angle)
			if (diff < 1.5) then
				if ((streetIdxToIdxAngle[streetIdxAngle] ~= nil and streetIdxToIdxAngle[streetIdxAngle] < diff) == false) then
					streetIdxToIdx[streetIdxAngle] = streetIdx
					streetIdxToIdxAngle[streetIdxAngle] = diff
					found = true
				end
			end
		end
		if (found == false) then
			print("loadConnecttionMap: removed streetIdxAngle: "..streetIdxAngle.." -> "..angle)
		end
	end

	for i, con in pairs(addCons) do
		local srcStreetIdx = streetIdxToIdx[con[1]] or nil
		local destStreetIdx = streetIdxToIdx[con[3]] or nil
		if (srcStreetIdx ~= nil and destStreetIdx ~= nil) then
			setTableValue(connectionMap, { srcStreetIdx, con[2], destStreetIdx, con[4] }, 1 )
		end
	end
	for i, con in pairs(removeCons) do
		local srcStreetIdx = streetIdxToIdx[con[1]] or nil
		local destStreetIdx = streetIdxToIdx[con[3]] or nil
		if (srcStreetIdx ~= nil and destStreetIdx ~= nil) then
			setTableValue(connectionMap, { srcStreetIdx, con[2], destStreetIdx, con[4] }, 2 )
		end
	end
end



function NodeWindow.stateNodeEditor:saveConnecttionMap(statedata)
	local addCons = {}
	local removeCons = {}
	local streetIdxToAngle = {}

	statedata.nodeData.addCons = addCons
	statedata.nodeData.removeCons = removeCons
	statedata.nodeData.streetIdxToAngle = streetIdxToAngle

	if (statedata.connectionMap ~= nil) then
		for srcStreetDataIdx, srcStreetData in pairs(statedata.connectionMap) do
			for srcLaneAnchorIdx, destStreetMap in pairs(srcStreetData) do
				for destStreetDataIdx, destStreetData in pairs(destStreetMap) do
					for destStreetAnchorIdx, status in pairs(destStreetData) do
						--print("("..srcStreetDataIdx.."-"..srcLaneAnchorIdx..") --> ("..destStreetDataIdx.."-"..destStreetAnchorIdx..") = "..status)
						if (status == 1) then
							addCons[#addCons+1] = { srcStreetDataIdx, srcLaneAnchorIdx, destStreetDataIdx, destStreetAnchorIdx }
						elseif (status == 2) then
							removeCons[#removeCons+1] = { srcStreetDataIdx, srcLaneAnchorIdx, destStreetDataIdx, destStreetAnchorIdx }
						end
					end
				end
			end
		end
	end

	for streetDataIdx, streetData in pairs(statedata.streetData) do
		streetIdxToAngle[streetDataIdx] = mathround(streetData.angle, 1)
	end
	nd.setNodeData(statedata.nodeData)
end


function NodeWindow.stateNodeEditor:getStateLaneAnchorByPos(pos, filterForward)
	local statedata = self
	for streetDataIdx, streetData in pairs(statedata.streetData) do
		for laneIdx, laneAnchor in pairs(streetData.laneAnchors) do
			local diff = (laneAnchor.pos[1]-pos[1])^2 + (laneAnchor.pos[2]-pos[2])^2
			if (diff < (laneAnchor.radius)^2) then
				if (filterForward ~= nil and (laneAnchor.forward ~= filterForward)) then
					-- skip if filter does not match
				else
					return streetDataIdx, laneAnchor.idx, laneAnchor
				end
			end
		end
	end
	return nil, nil
end

function NodeWindow.stateNodeEditor:changeNodeStatus(statedata, srcStreetDataIdx, srcLaneAnchorIdx, destStreetDataIdx, destLaneAnchorIdx)
	if (type(statedata.connectionMap[srcStreetDataIdx]) ~= "table") then
		statedata.connectionMap[srcStreetDataIdx] = {}
	end
	if (type(statedata.connectionMap[srcStreetDataIdx][srcLaneAnchorIdx]) ~= "table") then
		statedata.connectionMap[srcStreetDataIdx][srcLaneAnchorIdx] = {}
	end

	local srcMap = statedata.connectionMap[srcStreetDataIdx][srcLaneAnchorIdx]
	if (type(srcMap[destStreetDataIdx]) ~= "table") then
		srcMap[destStreetDataIdx] = {}
	end

	local status = srcMap[destStreetDataIdx][destLaneAnchorIdx] or 0
	if (status == 0) then
		status = 1
	elseif (status == 1) then
		status = 2
	elseif (status == 2) then
		status = nil
	end
	srcMap[destStreetDataIdx][destLaneAnchorIdx] = status
	statedata.saveConnecttionMap(self, statedata)
	local that = self
	local fn = function(success, message)
		self.lastRebuidError = message
		that:refresh()
	end
	self.lastRebuidError = nil
	nd.rebuildNode(statedata.nodeData.pos, fn, true)
end


function NodeWindow.stateNodeEditor:showMapMarkers(statedata, cameraPos)
	local nodeData = statedata.nodeData
	self:addCircleMarker(nodeData.id, nodeData.pos, { 0.0, 1.0, 0.0, 0.3}, (statedata.baseNodeRadius or 3) + 10, 9)

	if (statedata.streetData ~= nil) then
		for idx, streetData in pairs(statedata.streetData) do
--			self:addNodeMarker(nodeData.id.."-"..idx, streetData.xpos, { 1.0, 0.0, 1.0, 0.7}, 1)
			for idxLane, laneAnchor in pairs(streetData.laneAnchors) do
				if (laneAnchor.forward == false) then
					self:addNodeMarker(nodeData.id.."-"..idx.."..lane"..laneAnchor.idx, laneAnchor.pos, { 0.0, 0.7, 0.0, 0.7}, 1.5, 7)
				elseif (statedata.advMode == true ) then
					self:addNodeMarker(nodeData.id.."-"..idx.."..lane"..laneAnchor.idx, laneAnchor.pos, { 0.7, 0.0, 0.7, 0.7}, 1.5, 7)
				end
			end
		end
	end
end



NodeWindow.stateSelectLaneAnchor = {}

function NodeWindow.stateSelectLaneAnchor:showEditorPane(statedata, uilist)
	uilist[#uilist + 1] = {
		typ = "Label",
		caption = self._("Toggle lane node status by left clicking on node with a colored highlight, right click return to source node selection.")
	}
	local uiTableData = {}
	uilist[#uilist + 1] = {
		typ = "Table",
		id = "stateSelectLaneAnchorHelpTable",
		columns = 2,
		border = false,
		uilist = uiTableData,
		size = {
			0,
			-100
		},
		flags = {
			Sortable = false,
		},
	}

	local function createHelpItem(color, description, styleclass)
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = self._(color),
			column = 0,
			rownew = true,
		}
		if (styleclass ~= nil) then
			uiTableData[#uiTableData].styleclass = styleclass
		end
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = ""..self._(description),
			column = 1
		}
	end
	createHelpItem("Blue", "Already selected source node, to change return to source node selection by right click.", "commonapi-nodeeditor-blue")
	createHelpItem("Yellow", "No data set on destination yet, click to change.", "commonapi-nodeeditor-yellow")
	createHelpItem("Green", "Add edge lane connection between blue and green node.", "commonapi-nodeeditor-green")
	createHelpItem("Red", "Remove connection between blue and red node.", "commonapi-nodeeditor-red")

	if (self.lastRebuidError ~= nil) then
		uilist[#uilist + 1] = {
			typ = "Label",
			caption = self._("Last rebuild error:")
		}
		uilist[#uilist + 1] = {
			typ = "Label",
			caption = ""..self.lastRebuidError
		}
	end
	return uilist
end

function NodeWindow.stateSelectLaneAnchor:showMapMarkers(statedata, cameraPos)
	local nodeData = statedata.editor.nodeData
	local connectionMap = statedata.editor.connectionMap
	local stateStreetData = statedata.editor.streetData

	local srcStreetConMap = connectionMap[statedata.srcStreetDataIdx] or {}
	local srcStreetLaneConMap = srcStreetConMap[statedata.srcLaneAnchorIdx] or {}

	if (stateStreetData ~= nil) then
		for idx, streetData in pairs(stateStreetData) do
			local destStreetLaneConMap = srcStreetLaneConMap[idx] or {}
			for idxLane, laneAnchor in pairs(streetData.laneAnchors) do
				if (laneAnchor.forward ~= false or statedata.editor.advMode == true) then
					local status = destStreetLaneConMap[laneAnchor.idx] or 0
					if (status == 0) then
						self:addNodeMarker(nodeData.id.."-"..idx.."-lane"..laneAnchor.idx, laneAnchor.pos, { 1.0, 1.0, 0.0, 0.7}, 1.5, 7)
					elseif (status == 1) then
						self:addNodeMarker(nodeData.id.."-"..idx.."-lane"..laneAnchor.idx, laneAnchor.pos, { 0.0, 0.7, 0.0, 0.7}, 2.5, 7)
					elseif (status == 2) then
						self:addNodeMarker(nodeData.id.."-"..idx.."-lane"..laneAnchor.idx, laneAnchor.pos, { 0.7, 0.0, 0.0, 0.7}, 2.5, 7)
					end
				end
			end
		end
	end

	-- Show src node
	local srcLaneAnchorIdx = statedata.srcLaneAnchorIdx
	local streetData = stateStreetData[statedata.srcStreetDataIdx]
	for idxLane, laneAnchor in pairs(streetData.laneAnchors) do
		if (laneAnchor.idx == srcLaneAnchorIdx ) then
			self:addNodeMarker(nodeData.id.."-"..statedata.srcStreetDataIdx.."-lane"..laneAnchor.idx, laneAnchor.pos, { 0.0, 0.0, 0.5, 0.7}, 2.5, 7)
		end
	end
end

function NodeWindow.stateSelectLaneAnchor:clickRightOnMap(statedata, pos)
	local prevState = statedata.editor
	statedata.editor = nil
	self:selectState("stateNodeEditor", prevState)
end

function NodeWindow.stateSelectLaneAnchor:clickLeftOnMap(statedata, pos)
	local filterForward
	if (statedata.editor.advMode ~= true) then
		filterForward = true
	end
	local streetDataIdx, laneAnchorIdx, laneAnchor = statedata.editor:getStateLaneAnchorByPos(pos, filterForward)
	if (streetDataIdx ~= nil and (laneAnchor.forward ~= false or statedata.editor.advMode == true ) ) then
		--print("NodeWindow.stateSelectLaneAnchor:clickLeftOnMap", streetDataIdx)
		statedata.editor.changeNodeStatus(self, statedata.editor, statedata.srcStreetDataIdx, statedata.srcLaneAnchorIdx, streetDataIdx, laneAnchorIdx)
		--print("installShowMapMarkersCB / Save")
		self:installShowMapMarkersCB(true)
		self:refresh()
	end
end




-- Show MapMarkers
function NodeWindow:installShowMapMarkersCB(b)
	self._showMapMarkersCB = b
	if (self._showMapMarkersCB ~= true) then
		if (self._ostepconnection ~= nil) then
			local layout = self._ugwindow:getLayout()
			layout:removeItem(self._ostepconnection)
			api.gui.util.destroyLater(self._ostepconnection)
			self._ostepconnection = nil
			self._ostepconnectcampos = nil
			self._ostepconnectcnt = nil
		end
	else
		self._ostepconnectcampos = nil	-- reset
		self._ostepconnectcnt = nil
		local that = self;
		if (self._ostepconnection == nil) then
			local layout = self._ugwindow:getLayout()
			self._ostepconnection =  api.gui.comp.Component.new("TIMERELEMENT")
			layout:addItem(self._ostepconnection)
			self._ostepconnection:onStep(function()
				local x = that._ostepconnectcnt or 0
				x = x + 1
				if (x > 50) then
					x = 0
				end
				that._ostepconnectcnt = x
				if (x ~= 50) then
					return
				end

				local ok, err = pcall(that.showMapMarkersCB, that)
				if (ok == false) then
					print(err)
				end
			end)
		end
		local ok, err = pcall(that.showMapMarkersCB, that)
		if (ok == false) then
			print(err)
		end
	end
end

function NodeWindow:showMapMarkersCB()
	local cameraPos = game.gui.getCamera()
	local tmp = cameraPos[1] + cameraPos[2]

	if (self._ostepconnectcampos == nil or math.abs(tmp - self._ostepconnectcampos) > 10)  then
		self._ostepconnectcampos = tmp
		self:clearNodeMarkers()
		self:clearCircleMarkers()
		self:showMapMarkers(cameraPos)
	end
end
-- Show MapMarkers End


function NodeWindow:deleteNodeDataButton(nodeData)
	if (nodeData == nil and self.curstate ~= nil) then
		nodeData = self.curstate.nodeData or nil
	end
	if (nodeData == nil) then
		return
	end
	nd.deleteNodeData(nodeData)
	nd.rebuildNode(nodeData.pos, nil, true)
	self:selectState("stateSelectNode")
end

function NodeWindow:createUIListTabNodeEditor()
	local uilist = {}
	local uilistbuttonsHeader = {}

	uilist[#uilist + 1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttonsHeader
	}

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Button",
		id = "Refresh",
		caption = self._("Refresh"),
		onclick = function(item, win)
			win:refresh()
		end
	}

	local curRenderSettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
	self.params["showLanes"] = curRenderSettings["showLanes"]

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Checkbox",
		id = "showLanes",
		caption = self._("Show Lanes"),
		onclick = function(item, win)
			local cursettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
			cursettings["showLanes"] = not cursettings["showLanes"]
			win:refresh()
		end
	}

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Checkbox",
		id = "showDebugData",
		default = false,
		caption = self._("Show debug data"),
		onclick = function(item, win)
			win:refresh()
		end
	}

	local styleclassSelectNode

	if (self.curstate == nil) then
		styleclassSelectNode = "color-button"
	end
	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Button",
		id = "selectNode",
		default = false,
		caption = self._("Select base node by mouse"),
		styleclass = styleclassSelectNode,
		onclick = function(item, win)
			self:selectState("stateSelectNode")
		end
	}

	local ok, err = pcall(self.showEditorPane, self, uilist)
	if (ok == false) then
		print(err)
	end
	return uilist
end

return NodeWindow