local RegionWindow = {}

local ui = require("commonapi2.ui2.tpfui")
local inspectlua = require("commonapi2.extlib.inspect")

local zoneutil = require "mission.zone"
local hs = require("commonapi2.lineextradetails.helperscript")
local commonapimath = require("commonapi2._math")
local svec3 = commonapimath.svec3
local svec2 = commonapimath.svec2
local vec3 = commonapimath.vec3

local secondaryui = require "commonapi2.ui2.secondaryui"

require("serialize")
RegionWindow = ui.Window:new()

RegionWindow.id = "RegionWindow"
RegionWindow.caption = "Regions & Zones"
RegionWindow.uilist = {}

RegionWindow.height = 500
RegionWindow.width = 500
RegionWindow._zoneid = "commonapi2_RegionWindow_zone"

local function getSpeedConvert()
	local appcfg = api.util.getAppConfig()
	if (appcfg.speedUnit == 1) then
		return 0.44704
	end
	return 3.6
end
RegionWindow.speedFactor = getSpeedConvert()

local function mathround(num, deci)
	local d = 10 ^ (deci or 2)
	num = num * d
	if num >= 0 then
		num = math.floor(num + 0.5)
	else
		num = math.ceil(num - 0.5)
	end
	return num / d
end

function RegionWindow.create()
	local o
	o = RegionWindow:new()
	o:createUIList()
	o:open()
	o:setPositionCenter()

	return o
end

function RegionWindow:onclose()
	self:clearMakers()
	self:clearRegionMarkers()
end

function RegionWindow:refresh()
	self:createUIList()
	self:rebuild()
end


function RegionWindow:createUIList()
	local uilist = {}
	local uitabs = {}
--[[
	uitabs[#uitabs + 1] = {
		typ = "TabItem",
		id = "RegionTab",
		caption = "Entity",
		uilist = self:createUIListTabRegion()
	}

	uilist[#uilist + 1] = {
		typ = "TabBar",
		id = "RegionWindowTabs",
		uilist = uitabs
	}
--]]
	self.uilist = self:createUIListTabRegion()
end

function RegionWindow:addRegionMarker(id, pos, polygon, color)
	if (self._regionmarkers == nil) then
		self._regionmarkers = {}
	end
	if (color == nil) then
		color = {.8, .8, .8, .7}
	end
	local zonedata = {
		polygon = polygon,
		draw = true,
		drawColor = color
	}
	self._regionmarkers[id] = true
	game.interface.setZone(id, zonedata)
end

function RegionWindow:clearRegionMarkers()
	if (self._regionmarkers == nil) then
		return
	end
	for id, v in pairs(self._regionmarkers) do
		game.interface.setZone(id)
	end
	self._regionmarkers = nil
end

function RegionWindow:zoneMakerChange(id, pos, radius, color, polygon)
	if (self._zonemarkers == nil) then
		self._zonemarkers = {}
	end

	local olddata = self._zonemarkers[id]
	if (olddata ~= nil) then
		if (olddata.pos[1] == pos[1] and olddata.pos[2] == pos[2]) then
			return
		end
	end
	if (color == nil) then
		color = {.8, .8, .8, .7}
	end
	local newdata = {
		pos = {
			pos[1], pos[2]
		},
		radius = radius,
		color = color
	}
	self._zonemarkers[id] = newdata

	if (polygon == nil) then
		polygon = zoneutil.makeCircleZone(newdata.pos, newdata.radius, 16)
	end
	local zonedata = {
		polygon = polygon,
		draw = true,
		drawColor = newdata.color
	}
	game.interface.setZone(id, zonedata)
end

function RegionWindow:clearMaker(id)
	if (self._zonemarkers == nil) then
		return
	end
	self._zonemarkers[id] = nil
	game.interface.setZone(id)
end

function RegionWindow:clearMakers()
	if (self._zonemarkers == nil) then
		return
	end
	for id, v in pairs(self._zonemarkers) do
		game.interface.setZone(id)
	end
	self._zonemarkers = nil
end



function RegionWindow:fixPoly(poly)
	local sumX = 0
	local sumY = 0
	local cnt = #poly
    for i = 1, cnt do
		sumX = sumX + poly[i][1]
		sumY = sumY + poly[i][2]
	end

	local centerX =	sumX / cnt
	local centerY =	sumY /cnt
	local radius2 = 0
	for i = 1, cnt do
		local dx = poly[i][1] - centerX
		local dy = poly[i][2] - centerY
		poly[i][3] = dx
		poly[i][4] = dy

		local d = (dx * dx) + (dy * dy)
		if (d > radius2) then
			radius2 = d;
		end
	end

	local sortfn = function(a, b)
		if (a[3] >= 0 and b[3] < 0) then
			return true;
		end

		if (a[3] < 0 and b[3] >= 0) then
			return false;
		end

		if (a[3] == 0 and b[3] == 0) then
			if (a[4] >= 0 or b[4] >= 0) then
				return a[4] > b[4]
			end
			return b[4] > a[4]
		end

		local d = a[3] * b[4] - b[3] * a[4]
		if d < 0 then
			return true
		elseif d > 0 then
			return false
		end

		local l1 = a[3] * a[3] + a[4] * a[4]
		local l2 = b[3] * b[3] + b[4] * b[4]
		return l1 > l2
	end
	table.sort(poly, sortfn)

	for i = 1, #poly do
		poly[i][4] = nil
		poly[i][3] = nil
	end

	return { centerX, centerY }, radius2
end

function RegionWindow:isPointInPoly(poly, point)
	-- CrossingsMultiplyTest
	local yflag0, yflag1, inside_flag
	local vtx0, vtx1

	local numverts = #poly

	vtx0 = poly[numverts]
	vtx1 = poly[1]

	-- get test bit for above/below X axis
	yflag0 = ( vtx0[2] >= point[2] )
	inside_flag = false

	for i=2, numverts+1 do
		yflag1 = ( vtx1[2] >= point[2] )
		--[[ Check if endpoints straddle (are on opposite sides) of X axis
		 * (i.e. the Y's differ); if so, +X ray could intersect this edge.
		 * The old test also checked whether the endpoints are both to the
		 * right or to the left of the test point.  However, given the faster
		 * intersection point computation used below, this test was found to
		 * be a break-even proposition for most polygons and a loser for
		 * triangles (where 50% or more of the edges which survive this test
		 * will cross quadrants and so have to have the X intersection computed
		 * anyway).  I credit Joseph Samosky with inspiring me to try dropping
		 * the "both left or both right" part of my code.
		 --]]
		if ( yflag0 ~= yflag1 ) then
			--[[ Check intersection of pgon segment with +X ray.
			 * Note if >= point's X; if so, the ray hits it.
			 * The division operation is avoided for the ">=" test by checking
			 * the sign of the first vertex wrto the test point; idea inspired
			 * by Joseph Samosky's and Mark Haigh-Hutchinson's different
			 * polygon inclusion tests.
			 --]]
			if ( ((vtx1[2] - point[2]) * (vtx0[1] - vtx1[1]) >= (vtx1[1] - point[1]) * (vtx0[2] - vtx1[2])) == yflag1 ) then
				inside_flag = not inside_flag
			end
		end

		-- Move to the next pair of vertices, retaining info as possible.
		yflag0  = yflag1
		vtx0    = vtx1
		vtx1    = poly[i]
	end

	return  inside_flag
end


local regionData = {}

--- [[
function RegionWindow:getRegionById(id)
	local region
	if (id == nil) then
		region = commonapi._native.FlexStreets.createRegion()
	else
		region = commonapi._native.FlexStreets.getRegionById(id)
		if (region ~= nil) then
			region.id = id
		end
	end
	return region
end

function RegionWindow:getRegions()
	return commonapi._native.FlexStreets.getRegions()
end

function RegionWindow:getRegionsByPosition(x, y)
	return commonapi._native.FlexStreets.getRegionsByPos(x, y)
end



function RegionWindow:setRegion(region)
	commonapi._native.FlexStreets.setRegionById(region.id, region)
end

function RegionWindow:deleteRegion(region)
	if (region.id ~= nil) then
		commonapi._native.FlexStreets.setRegionById(region.id, nil)
	end
end

function RegionWindow:deleteRegionById(regionid)
	if (regionid ~= nil) then
		commonapi._native.FlexStreets.setRegionById(regionid, nil)
	end
end
--]]

--[[
function RegionWindow:getRegionById(id)
	return regionData[id] or nil
end

function RegionWindow:getRegions()
	return regionData
end


function RegionWindow:setRegion(region)
	regionData[region.id] = region
end

function RegionWindow:deleteRegion(region)
	if (region.id ~= nil) then
		regionData[region.id] = nil
	end
end
--]]






function RegionWindow:addOrSetRegion(region)
	if (region.id == nil) then
		math.randomseed(os.time())
		region.id = "Region_"..math.random(0, 100000)
	end
	self:setRegion(region)
end


function RegionWindow:saveRegion(region)
	if (region.id == nil) then
		math.randomseed(os.time())
		region.id = "Region_"..math.random(0, 100000)
	end
	if (region.center == nil) then
		return
	end
	if (#region.polygon < 3) then
		return
	end
	self:setRegion(region)
end


function RegionWindow:editRegionPolygon(id)
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end
	self._currentRegion = self:getRegionById(id)
	if (self._currentRegion == nil) then
		print("Creating new region")
		self._currentRegion = {
			polygon = {},
		}
	end
	if (id == nil) then
		self._currentRegion.center = nil
	end
	local win = self
	secondaryui.setMouseListener({
		filterui = false,
		fn = function(evt, mouseScreenPos)
			return win:onAddPolygonMouseEvent(evt, mouseScreenPos)
		end,
		cleanup = function()
			local ok, err = pcall(win.onAddPolygonMouseEvent_cleanup, win)
			if (ok == false) then
				print(err)
			end
		end,
	})
	self:onAddPolygonMouseEvent_renderMarkers(self._currentRegion.center or nil, self._currentRegion.polygon)

	if (self._currentRegion.center ~= nil) then
		local cameraPos = game.gui.getCamera()
		cameraPos[1] = self._currentRegion.center[1]
		cameraPos[2] = self._currentRegion.center[2]
		game.gui.setCamera(cameraPos)
	end
end


function RegionWindow:onAddPolygonMouseEvent_cleanup()
	self:clearMakers()
	self.params["RegionStartNewRegion"] = false
	self:refresh()
end


function RegionWindow:onAddPolygonMouseEvent_renderMarkers(pos, poly)
	if (pos == nil) then
		return
	end
	local color = {.0, .0, 1.0, .7}
	if (#poly > 2) then
		self:zoneMakerChange("curRegion", pos, 5, color, poly)
		self:clearMaker("curRegionPoint2")
	else
		self:clearMaker("curRegion")
		if (poly[1] ~= nil) then
			self:zoneMakerChange("curRegion", poly[1], 2, color)
		end
		if (poly[2] ~= nil) then
			self:zoneMakerChange("curRegionPoint2", poly[2], 2, color)
		end
	end
end


function RegionWindow:onAddPolygonMouseEvent(evt, mouseScreenPos)
	if (self._ugwindow == nil) then
		secondaryui.setMouseListener(nil)
		return
	end
	if ( (evt.type == 2 and evt.button == 2) or (evt.type == 7)) then
		local regionToSave = self._currentRegion
		secondaryui.setMouseListener(nil)
		self:saveRegion(regionToSave)
		self.params.curregion = regionToSave.id
		self:refresh()
		return
	end


	if (evt.type == 2 and evt.button == 1) then
		secondaryui.setMouseListener(nil)
		self:refresh()
		return
	end
	local _ugwindow = self._ugwindow
	local rect = _ugwindow:getContentRect()


	if (self._currentRegion == nil) then
		secondaryui.setMouseListener(nil)
		return
	end
	local _currentRegion = self._currentRegion


	if (rect:contains(mouseScreenPos.x, mouseScreenPos.y)) then
		return
	end
	local pos = game.gui.getTerrainPos()
	if (pos == nil) then
		return
	end

	local poly = _currentRegion.polygon
	local nearpoint = -1
	for idx, p in ipairs(poly) do
		local length = (pos[1]-p[1])^2 + (pos[2]-p[2])^2

		if (length < 10) then
			nearpoint = idx
		end
	end

	local colorH = {1.0, .0, 0.0, .7}
	local color2 = {.0, 1.0, 1.0, .7}

	if (nearpoint > -1) then
		self:zoneMakerChange("selecthighlight", pos, 5, color2)
	else
		self:clearMaker("selecthighlight")
	end


	if (evt.type == 2 and evt.button == 0) then
		if (nearpoint > -1) then
			table.remove(poly, nearpoint)
		else
			local rpoint = { pos[1], pos[2] }
			poly[#poly + 1] = rpoint
			print("Added new polypoint: ("..rpoint[1] ..", "..rpoint[2]..")")
		end
		local center, radius2 = self:fixPoly(poly)
		_currentRegion.center = center
		_currentRegion.radius2 = radius2
		self:onAddPolygonMouseEvent_renderMarkers(pos, poly)
		--secondaryui.setMouseListener(nil)
		--self:refresh()
		return true
	end

	local cursorcolor = {.8, .8, .8, .7}

	if (#poly > 2 and self:isPointInPoly(poly, pos)) then
		cursorcolor = {.8, .8, 0, .7}
	end

	self:zoneMakerChange("inspectcursor", pos, 10, cursorcolor)


	if (evt.type < 3 and evt.button == 0) then
		return true
	end
end


function RegionWindow:installSelectRegion()
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end

	local win = self
	secondaryui.setMouseListener({
		filterui = false,
		fn = function(evt, mouseScreenPos)
			--return win:onFindPolygonMouseEvent(evt, mouseScreenPos)
			local ok, result = pcall(win.onFindPolygonMouseEvent, win, evt, mouseScreenPos)
			if (ok == false) then
				print(result)
				return false
			end
			return result
		end,
		cleanup = function()
			local ok, err = pcall(win.onFindPolygonMouseEvent_cleanup, win)
			if (ok == false) then
				print(err)
			end
		end,
	})

	if (self.params.showRegions ~= true) then
		self.params.showRegions = true
		self:installShowRegionsCB()
	end
end


function RegionWindow:onFindPolygonMouseEvent_cleanup()
	self.params.selectRegion = false
	self:refresh()
end

function RegionWindow:onFindPolygonMouseEvent(evt, mouseScreenPos)
	if (self._ugwindow == nil) then
		secondaryui.setMouseListener(nil)
		return
	end
	if ( (evt.type == 2 and evt.button == 2) or (evt.type == 7)) then
		secondaryui.setMouseListener(nil)
		self:refresh()
		return
	end

	if (self.params.selectRegion ~= true) then
		return
	end

	if (evt.type == 2 and evt.button == 1) then
		secondaryui.setMouseListener(nil)
		self:refresh()
		return
	end
	local _ugwindow = self._ugwindow
	local rect = _ugwindow:getContentRect()
	if (rect:contains(mouseScreenPos.x, mouseScreenPos.y)) then
		return
	end

	if (evt.type == 2 and evt.button == 0) then
		local pos = game.gui.getTerrainPos()
		if (pos == nil) then
			return
		end
		local regions = self:getRegionsByPosition(pos[1], pos[2])
		local firstRegionIndex, firstRegion = next(regions)
		if (firstRegionIndex) then
			local curregionid = self.params.curregion
			local prevregionid
			local newregionid
			for k, region in pairs(regions) do
				if (curregionid == region.id) then
					prevregionid = curregionid
				else
					if (prevregionid ~= nil) then
						newregionid = region.id
					end
				end
			end
			if (newregionid == nil or prevregionid == nil) then
				newregionid = firstRegion.id
			end
			print("Select region:", newregionid)
			self.params.curregion = newregionid
			self:showRegions()
			self:refresh()
		end
		return true
	end
end


function RegionWindow:installShowRegionsCB()
	if (self.params.showRegions ~= true) then
		if (self._ostepconnection ~= nil) then
			local layout = self._ugwindow:getLayout()
			layout:removeItem(self._ostepconnection)
			api.gui.util.destroyLater(self._ostepconnection)
			self._ostepconnection = nil
		end
	else
		if (self._ostepconnection == nil) then
			local layout = self._ugwindow:getLayout()
			self._ostepconnection =  api.gui.comp.Component.new("TIMERELEMENT")
			layout:addItem(self._ostepconnection)
			local that = self;
			self._ostepconnection:onStep(function()
				that:showRegionsCB()
			end)
		end
	end
end

function RegionWindow:showRegionsCB()
	local x = self._ostepconnectcnt or 0
	x = x + 1
	if (x > 50) then
		x = 0
	end
	self._ostepconnectcnt = x
	if (x ~= 50) then
		return
	end

	local cameraPos = game.gui.getCamera()
	local tmp = cameraPos[1] + cameraPos[2]

	if (self._ostepconnectcampos == nil or math.abs(tmp - self._ostepconnectcampos) > 10)  then
		self._ostepconnectcampos = tmp
		self:showRegions()
	end
end

function RegionWindow:showRegions()
	self:clearRegionMarkers()
	if (self.params.curregion ~= nil) then
		local region = self:getRegionById(self.params.curregion)
		if (region ~= nil) then
			self:addRegionMarker(region.id, region.center, region.polygon, { 0.0, 1.0, 0.0, 0.7})
		end
	end

	if (self.params.showRegions ~= true) then
		return
	end
	local cameraPos = game.gui.getCamera()
	local regions = self:getRegions()
	local l = 300 ^ 2
	for k, region in pairs(regions) do
		if (self.params.curregion == region.id) then
			-- pass
		else

			local c = region.center
			local l2 = ( (cameraPos[1]-c[1])^2 + (cameraPos[2]-c[2])^2 )
			if (l2 < l) then
				self:addRegionMarker(region.id, c, region.polygon)
			end
		end


	end

end




function RegionWindow:applyRebuildRegion()
	local regionId = self.params.curregion or nil
	if (regionId == nil) then
		return
	end

	local region = self:getRegionById(regionId)
	if (region == nil) then
		return
	end
	local centerx = region.center[1]
	local centery = region.center[2]
	local r2 = region.radius2
--[[
	local compBaseEdgeStreet = api.engine.getComponent(entityid, api.type.ComponentType.BASE_EDGE_STREET)
	local compBaseEdgeTrack = api.engine.getComponent(entityid, api.type.ComponentType.BASE_EDGE_TRACK)

	local compNode0 = api.engine.getComponent(compBaseEdge.node0, api.type.ComponentType.BASE_NODE)
	local compNode1 = api.engine.getComponent(compBaseEdge.node1, api.type.ComponentType.BASE_NODE)

	local compNode0 = api.engine.getComponent(compBaseEdge.node0, api.type.ComponentType.BASE_NODE)
	local compNode1 = api.engine.getComponent(compBaseEdge.node1, api.type.ComponentType.BASE_NODE)
--]]

	local checknode = function(compNode)
		local dx = compNode.position.x - centerx
		local dy = compNode.position.y - centery
		local dl2 = (dx * dx) + (dy * dy)
		return (dl2 < r2)
	end
	local proposal = api.type.SimpleProposal.new()
	local idx = 1
	local entities = {}
	local conEntities = {}
	local fn = function(entity, compBaseEdge)

		local compNode0 = api.engine.getComponent(compBaseEdge.node0, api.type.ComponentType.BASE_NODE)
		if (checknode(compNode0)) then
			local compNode1 = api.engine.getComponent(compBaseEdge.node0, api.type.ComponentType.BASE_NODE)
			if (checknode(compNode1)) then
				local conEntity = api.engine.system.streetConnectorSystem.getConstructionEntityForEdge(entity)
				if ( conEntity > -1 ) then
					conEntities[tostring(conEntity)] = conEntity
				else
					entities[#entities+1] = entity
					local baseEdge = api.engine.getComponent(entity, api.type.ComponentType.BASE_EDGE)
					local baseEdgeStreet = api.engine.getComponent(entity, api.type.ComponentType.BASE_EDGE_STREET)
					local baseEdgeTrack = api.engine.getComponent(entity, api.type.ComponentType.BASE_EDGE_TRACK)
					local eo = api.type.SegmentAndEntity.new()
					eo.entity = -idx
					eo.type = 0
					eo.comp = baseEdge
					if (baseEdgeStreet ~= nil) then
						local playerOwned = api.engine.getComponent(entity, api.type.ComponentType.PLAYER_OWNED)
						if (playerOwned ~= nil) then
							local playerOwnedComponent = api.type.PlayerOwned.new()
							playerOwnedComponent.player = playerOwned.player
							eo.playerOwned = playerOwnedComponent
						end
						eo.streetEdge = baseEdgeStreet
						eo.type = 0
					end
					if (baseEdgeTrack ~= nil) then
						eo.type = 1
						eo.trackEdge = baseEdgeTrack
					end
					proposal.streetProposal.edgesToRemove[idx] = entity
					proposal.streetProposal.edgesToAdd[idx] = eo
					idx = idx + 1
				end
			end  -- checknode checknode
		end
	end
	api.engine.forEachEntityWithComponent(fn, api.type.ComponentType.BASE_EDGE)

	local ctx2 = api.type.Context:new()
	ctx2.checkTerrainAlignment = false
	ctx2.cleanupStreetGraph = true
	ctx2.gatherBuildings = true
	ctx2.gatherFields = true

	local ignoreErrors = false
	if (self.params["rebuildIgnoreErrors"] == true) then
		ignoreErrors = true
	end

	local cmd = api.cmd.make.buildProposal(proposal, ctx2, ignoreErrors)
	local that = self
	local fnCmdErr = function(result, success)
		if (result ~= nil and result.resultProposalData ~= nil and result.resultProposalData.errorState) then
			local errorState = result.resultProposalData.errorState
			if (#errorState.messages > 0) then
				local messages = {}
				for j, msg in ipairs(errorState.messages) do
					messages[#messages+1] = msg
				end
				local msg = table.concat(messages, "\n")
				that.lastRebuildError = msg
				that:refresh()
				return
			end

			if (#errorState.warnings > 0) then
				local messages = {}
				for j, msg in ipairs(errorState.warnings) do
					messages[#messages+1] = msg
				end
				local msg = table.concat(messages, "\n")
				that.lastRebuildError = msg
				that:refresh()
				return
			end


			if (errorState.critical == true) then
				that.lastRebuildError = "Critical Error found"
				that:refresh()
				return
			end
		end

		that.lastRebuildError = nil
		that:refresh()
	end
	api.cmd.sendCommand(cmd, fnCmdErr)


	local fnRebuildCons = function(entityId)
		local proposalCon = api.type.SimpleProposal.new()
		print("Rebuilding construction by user request via regionwindow: ", entityId)
		local conold = api.engine.getComponent(entityId, api.type.ComponentType.CONSTRUCTION)
		local connew = api.type.SimpleProposal.ConstructionEntity.new()
		local params = hs.deepcopy(conold.params)
		connew.params = params
		connew.fileName = conold.fileName
		connew.transf = hs.deepcopy(conold.transf)

		local playerComponent = api.engine.getComponent(entityId, api.type.ComponentType.PLAYER_OWNED)
		if (playerComponent ~= nil) then
			connew.playerEntity = playerComponent.player
		else
			connew.playerEntity = -1
		end

		local nameComponent = api.engine.getComponent(entityId, api.type.ComponentType.NAME)
		if (nameComponent ~= nil) then
			connew.name = nameComponent.name
		else
			connew.name = ""
		end
		proposalCon.constructionsToAdd[1] = connew
		proposalCon.constructionsToRemove = { entityId }
	--	proposalCon.old2new[entityId] = -1

		local ctx = api.type.Context:new()
		ctx.checkTerrainAlignment = false
		ctx.cleanupStreetGraph = true
		ctx.gatherBuildings = true
		ctx.gatherFields = true
		local cmd = api.cmd.make.buildProposal(proposalCon, ctx, true)
		api.cmd.sendCommand(cmd, function(cbcmd, cbstatus)
			if cbcmd and cbcmd.resultProposalData and cbcmd.resultProposalData then
				commonapi.dmp(cbcmd.resultProposalData.errorState)
			end
		end)
	end

	-- commonapi.dmp(conEntities)
	for k, entityId in pairs(conEntities) do
		fnRebuildCons(entityId)
	end


end


function RegionWindow:editRegion(regionId)
	if (secondaryui.hasMouseListener()) then
		secondaryui.setMouseListener(nil)
	end

	if (self:getRegionById(regionId) ~= nil) then
		self.params.curregion = regionId
	else
		self.params.curregion = nil
	end
	self:refresh()
end


function RegionWindow:deleteRegionButton(regionid)
	self:deleteRegionById(regionid)
	self.params.curregion = nil
	self._currentRegion = nil
	self:refresh()
end

function RegionWindow:saveEditorRegion()
	local regionId = self.params.curregion or nil
	if (regionId == nil) then
		return
	end

	local region = self:getRegionById(regionId)
	if (region == nil) then
		return
	end
	local strMaxTrackSpeed = self.params["maxTrackSpeed"]
	region.maxTrackSpeed = (tonumber(strMaxTrackSpeed) or 0) / self.speedFactor
	if (region.maxTrackSpeed < 2.6) then
		region.maxTrackSpeed = 0;
	end

	local attributes = self.params["attributes"] or ""
	region.attributes = {}
	for attr in string.gmatch(attributes, "[^%s%c,\n\r\t]+") do
		region.attributes[#region.attributes+1] = attr
	end


	local disableParcels = self.params["disableParcels"] or 0
	if (disableParcels > 0) then
		region.disableParcels = disableParcels
	else
		region.disableParcels = nil
	end

	local strTrackCatenaryId = self.params["trackCatenaryId"]
	local trackCatenaryId =(tonumber(strTrackCatenaryId) or -1)
	if (trackCatenaryId >= 0) then
		region.trackCatenaryId = trackCatenaryId
	else
		region.trackCatenaryId = nil
	end
	self:setRegion(region)
end

function RegionWindow:showEditorRegion(uilistparent)
	local regionId = self.params.curregion or nil
	if (regionId == nil) then
		return
	end

	local region = self:getRegionById(regionId)
	if (region == nil) then
		uilistparent[#uilistparent+1] = {
			typ = "Label",
			caption = "Region not found with: "..regionId
		}
		return
	end

	local maxSpeed = (region.maxTrackSpeed or 0) * self.speedFactor
	self.params["maxTrackSpeed"] = ""..mathround(maxSpeed, 1)
	self.params["attributes"] = table.concat( (region.attributes or {}), ", ")
	self.params["disableParcels"] =  region.disableParcels or 0

	local trackCatenaryId = region.trackCatenaryId or -1
	if (trackCatenaryId == nil or trackCatenaryId == -1) then
		self.params["trackCatenaryId"] = nil
	else
		self.params["trackCatenaryId"] = ""..trackCatenaryId
	end

	local cursettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
	self.params["showParcels"] = not cursettings["hideParcels"]

	uilistparent[#uilistparent+1] = {
		typ = "Label",
		caption = "Region: "..regionId
	}

	local uiparams = {}

	uiparams[#uiparams + 1] = {
		typ = "Label",
		caption = "Max (rail) speed:",
		column = 0,
		rownew = true
	}
	uiparams[#uiparams + 1] = {
		typ = "TextInput",
		id = "maxTrackSpeed",
		column = 1,
	}

	uiparams[#uiparams + 1] = {
		typ = "Label",
		caption = "Railtrack catenary id:",
		column = 0,
		rownew = true
	}

	uiparams[#uiparams + 1] = {
		typ = "TextInput",
		id = "trackCatenaryId",
		column = 1,
	}

	uiparams[#uiparams + 1] = {
		typ = "Label",
		caption = "Disable Parcels:",
		column = 0,
		rownew = true
	}
	uiparams[#uiparams + 1] = {
		typ = "ComboBox",
		id = "disableParcels",
		values = { "None", "Left", "Right", "Both Sides" },
		defaultIndex = 0,
		column = 1,
	}

	uiparams[#uiparams + 1] = {
		typ = "Checkbox",
		id = "showParcels",
		caption = "Show Parcels",
		column = 2,
		onclick = function(item, win)
			local cursettings = commonapi._native.UIHacks.getRenderSettings( api.gui.util.getGameUI() )
			cursettings["hideParcels"] = not cursettings["hideParcels"]
			win:refresh()
		end
	}


	uiparams[#uiparams + 1] = {
		typ = "Label",
		caption = "Attributes:",
		column = 0,
		rownew = true
	}

	uiparams[#uiparams + 1] = {
		typ = "TextInput",
		id = "attributes",
		column = 1,
	}

	uilistparent[#uilistparent+1] = {
		typ = "Table",
		id = "RegionSetting",
		columns = 3,
		columnWidths = { 100, 120, 100},
		border = false,
		uilist = uiparams,
	}

	local uilistsaveapply = {}
	uilistparent[#uilistparent + 1] = {
		typ = "LayoutHorizontal",
		uilist = uilistsaveapply
	}

	uilistsaveapply[#uilistsaveapply+1] = {
		typ = "Button",
		id = "saveRegionSetting",
		caption = self._("Save"),
		onclick = function(item, win)
			win:saveEditorRegion()
			win:refresh()
		end,
	}

	uilistsaveapply[#uilistsaveapply+1] = {
		typ = "Button",
		id = "applyRegionSetting",
		caption = self._("Apply").." (Rebuilds streets, tracks in region)",
		onclick = function(item, win)
			win:saveEditorRegion()
			win:applyRebuildRegion()
			win:refresh()
		end,
	}

	uilistsaveapply[#uilistsaveapply+1] = {
		typ = "Checkbox",
		id = "rebuildIgnoreErrors",
		caption = self._("Ignore Errors"),
		onclick = function(item, win)
			win:refresh()
		end,
	}

	if (self.lastRebuildError ~= nil) then
		uilistparent[#uilistparent + 1] = {
			typ = "Label",
			caption = self._("Last rebuild error:").." "..(self.lastRebuildError or "")
		}
	end
end


function RegionWindow:createUIListTabRegion()
	local uilist = {}
	local uilistbuttonsHeader = {}

	uilist[#uilist + 1] = {
		typ = "LayoutHorizontal",
		uilist = uilistbuttonsHeader
	}

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "TextInput",
		id = "regionId",
	}

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Button",
		id = "Refresh",
		caption = "Refresh",
		onclick = function(item, win)
			win:refresh()
		end
	}

	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "Checkbox",
		id = "showRegions",
		default = false,
		caption = self._("Show nearby regions"),
		onclick = function(item, win)
			win:installShowRegionsCB()
			win:refresh()
		end
	}


	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "ToggleButton",
		id = "selectRegion",
		default = false,
		caption = self._("Select region by mouse"),
		onclick = function(item, win)
			win:installSelectRegion()
			win:refresh()
		end
	}


	uilistbuttonsHeader[#uilistbuttonsHeader + 1] = {
		typ = "ToggleButton",
		id = "RegionStartNewRegion",
		caption = self._("Create new region by mouse"),
		default = false,
		onclick = function(item, win)
			if (win.params["RegionStartNewRegion"] == true) then
				win:editRegionPolygon(nil)
				win:refresh()
			else
				if (secondaryui.hasMouseListener()) then
					secondaryui.setMouseListener(nil)
				end
			end
		end
	}
	if (self.params["RegionStartNewRegion"] == true) then
		uilist[#uilist + 1] = {
			typ = "Label",
			caption = "Add atleast three points, left click for adding polygon points, right click exit mode.\nClick on point to remove it.",
		}
	end

	self:showEditorRegion(uilist)



	local uiTableData = {}
	local headerColumns = {
		{
			caption = "RegionId",
			width = 50
		},
		{
			caption = _("Location"),
			width = 50
		},
		{
			caption = ""
		},
		{
			caption = ""
		},
		{
			caption = ""
		}
	}
	local TableHeader = {
		typ = "TableHeader",
		items = headerColumns,
	}
	uiTableData[#uiTableData + 1] = TableHeader

	local regionlist2id = {}

	uilist[#uilist + 1] = {
		typ = "Table",
		id = "RegionsTable",
		columns = #headerColumns,
		border = false,
		uilist = uiTableData,
		size = {
			0,
			-100
		},
		flags = {
			Sortable = false,
			MultiSortable = false,
			ScrollFreezeTopRow = true
		},
		onclick = function(item, win)
			local index = win.params["RegionsTable"]
			local regions = win:getRegions()
			if (index >= 0 and regions) then
				--local entityid = self._zoneEntities[index + 1]
				--if (entityid) then
				--	win.params.EntityId = tostring(entityid)
				--end
			end
			win:refresh()
		end
	}
	local function createItem(region)
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = region.id,
			column = 0,
			rownew = true,
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Label",
			caption = "(x: "..region.center[1] .. "y: "..region.center[2]..")",
			column = 1
		}
		uiTableData[#uiTableData + 1] = {
			typ = "Button",
			caption = "Edit polygon",
			id = "RegionTable-editpoly-"..region.id,
			data = {
				regionId = region.id,
			},
			onclick = function(item, win)
				win:editRegionPolygon(item.data.regionId)
			end,
			column = 2
		}

		uiTableData[#uiTableData + 1] = {
			typ = "Button",
			caption = "Edit",
			id = "RegionTable-edit-"..region.id,
			data = {
				regionId = region.id,
			},
			onclick = function(item, win)
				win:editRegion(item.data.regionId)
			end,
			column = 3
		}

		uiTableData[#uiTableData + 1] = {
			typ = "Button",
			caption = "X",
			id = "RegionTable-delete-"..region.id,
			data = {
				regionId = region.id,
			},
			onclick = function(item, win)
				win:deleteRegionButton(item.data.regionId)
			end,
			column = 4
		}
	end

	local regions = self:getRegions()
	for id, region in pairs(regions) do
		createItem(region)
	end


	self:showRegions()

	return uilist
end

return RegionWindow